/*
madman - a music manager
Copyright (C) 2003  Andreas Kloeckner <ak@ixion.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/




#include <qdatetime.h>
#include <stdexcept>
#include "utility/base.h"
#include "utility/mt_random.h"
#include "database/auto_dj.h"




namespace
{
  mtRand RNG;
  bool RNGInitialized = false;

  unsigned randomInRange( unsigned max_plus_1 )
  {
    if ( max_plus_1 == 0 )
      throw runtime_error( "Invalid range for RNG" );

    if ( !RNGInitialized )
    {
      RNG.sgenrand( time( NULL ) );
      RNGInitialized = true;
    }

    unsigned result;
    // This should almost never loop, it catches MT returning 1.
    do
    {
      result = unsigned( RNG.gendrand() * max_plus_1 );
    }
    while ( result == max_plus_1 );

    return result;
  }

}








class tDecisionGenerator
{
    struct tAlternative
    {
      unsigned 	Ratio;
      string	DecisionId;
    };
    typedef vector<tAlternative>	tAlternativeList;
    tAlternativeList			Alternatives;

  public:
    void addAlternative( unsigned ratio, const string &decision_id )
    {
      tAlternative a = { ratio, decision_id };
      Alternatives.push_back( a );
    }

    string decide()
    {
      unsigned sum = 0;
      FOREACH_CONST( first, Alternatives, tAlternativeList )
	sum += first->Ratio;

      unsigned decision = randomInRange( sum );

      FOREACH_CONST( first, Alternatives, tAlternativeList )
      {
	if ( decision > first->Ratio )
	  decision -= first->Ratio;
	else
	  return first->DecisionId;
      }
      throw runtime_error( "Invalid RNG result" );
    }
};




namespace
{
  bool addNonduplicateSong( tSongList &list, tSong *song )
  {
    FOREACH_CONST( first, list, tSongList )
      if ( *first == song )
	return false;
    list.push_back( song );
    return true;
  }

  void makeList( tSongCollection &collection, tCriterion *crit, tSongList &result, 
      bool sort = false, tSongField sort_by = FIELD_ARTIST )
  {
    tSongSet set;
    set.setSongCollection( &collection );
    set.setCriterion( crit );
    set.reevaluateCriterion();
    if ( sort )
      set.sortBy( sort_by );

    result.clear();
    set.render( result );
  }

  void makeList( tSongCollection &collection, const QString &criterion, tSongList &result, 
      bool sort = false, tSongField sort_by = FIELD_ARTIST )
  {
    auto_ptr<tCriterion> parsed_crit( parseCriterion( criterion, false ) );
    makeList( collection, parsed_crit.release(), result, sort, sort_by );
  }

  void makeListWithBase( tSongCollection &collection, tSongSet *base, const QString &criterion, tSongList &result, 
      bool sort = false, tSongField sort_by = FIELD_ARTIST )
  {
    auto_ptr<tCriterion> parsed_crit( parseCriterion( criterion, false ) );
    if ( !base )
      makeList( collection, parsed_crit.release(), result, sort, sort_by );
    else
      makeList( collection, 
	  new tAndCriterion( new tInSetCriterion( base ), parsed_crit.release() ),
	  result, sort, sort_by );
  }
}




void doAutoDJ( tSongCollection &collection, tSongSet *base, tSongList &result, 
    tAutoDJPreferences &prefs, unsigned count )
{
  // since this is a randomized algorithm and it might never finish,
  // we need to bound run time.

  QTime start_time;
  start_time.start();

  result.clear();

  tDecisionGenerator decider;
  decider.addAlternative( prefs.PlayRatioUnrated, "unrated" );
  decider.addAlternative( prefs.PlayRatio5Rated, "5_rated" );
  decider.addAlternative( prefs.PlayRatio4Rated, "4_rated" );
  decider.addAlternative( prefs.PlayRatio3Rated, "3_rated" );
  decider.addAlternative( prefs.PlayRatio2Rated, "2_rated" );
  decider.addAlternative( prefs.PlayRatio1Rated, "1_rated" );
  decider.addAlternative( prefs.PlayRatio0Rated, "0_rated" );
  decider.addAlternative( prefs.PlayRatioUnplayed, "unplayed" );
  decider.addAlternative( prefs.PlayRatioOftenPlayed, "often_played" );
  decider.addAlternative( prefs.PlayRatioRarelyPlayed, "rarely_played" );
  decider.addAlternative( prefs.PlayRatioRandom, "random" );
  decider.addAlternative( prefs.PlayRatioUnratedByGoodArtist, "unrated_by_good_artist" );

  /*
  unsigned often_played_count;
  {
    tSongList list;
    makeList( collection, "~all", list, true, FIELD_PLAYCOUNT );

    if ( list.size() )
      often_played_count = list[ list.size() * 9 / 10 ]->playCount();
    else
      often_played_count = 0;
  }
  */
  
  tSongList list_all, list_unrated;
  makeListWithBase( collection, base, "~all", list_all );
  makeListWithBase( collection, base, "~unrated", list_unrated );

  if ( list_all.size() == 0 )
    throw runtime_error( "Empty base set for Auto DJ" );

  while ( result.size() < count )
  {
    if ( start_time.elapsed() > 5000 )
      throw runtime_error( "Auto DJ did not finish in time" );

    string decision = decider.decide();

    if ( decision == "unrated_by_good_artist" )
    {
      if ( list_unrated.size() == 0 )
	continue;

      while ( true )
      {
	if ( start_time.elapsed() > 5000 )
	  throw runtime_error( "Auto DJ did not finish in time" );

	unsigned index = randomInRange( list_unrated.size() );
	tSong *song = list_unrated[ index ];

	tSongList artist_list;
	QString criterion = QString( "~artist(complete:%1)" ).arg( 
	      quoteString( song->artist().lower() ) );
	makeList( collection, criterion, artist_list );

	int best_rating = -1;
	FOREACH_CONST( first, artist_list, tSongList )
	  if ( (*first)->rating() > best_rating )
	    best_rating = (*first)->rating();

	if ( best_rating >= 4 )
	{
	  addNonduplicateSong( result, song );
	  break;
	}
      }
      continue;
    }

    QString criterion;

    if ( decision == "unrated" )
      criterion = "~unrated";
    else if ( decision == "5_rated" )
      criterion = "~rating(5)";
    else if ( decision == "4_rated" )
      criterion = "~rating(4)";
    else if ( decision == "3_rated" )
      criterion = "~rating(3)";
    else if ( decision == "2_rated" )
      criterion = "~rating(2)";
    else if ( decision == "1_rated" )
      criterion = "~rating(1)";
    else if ( decision == "0_rated" )
      criterion = "~rating(0)";
    else if ( decision == "unplayed" )
      criterion = "~playcount(0)&~unrated";
    else if ( decision == "often_played" )
      criterion = "~playcount(>=3)";
    else if ( decision == "rarely_played" )
      criterion = "~playcount(<=2)&~playcount(>=1)";
    else if ( decision == "random" )
      criterion = "~all";
    else
      throw runtime_error( "Invalid decision generator result" );

    tSongList list;
    makeListWithBase( collection, base, criterion, list );

    if ( list.size() )
      addNonduplicateSong( result, list[ randomInRange( list.size() ) ] );
  }
}
