/*
madman - a music manager
Copyright (C) 2003  Andreas Kloeckner <ak@ixion.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/




#include <memory>
#include <iomanip>
#include <stdexcept>
#include <qregexp.h>
#include <qfileinfo.h>
#include <id3/tag.h>
#include <vorbis/vorbisfile.h>
#include <glib.h>
#include "utility/vcedit.h"
#include "utility/mp3tech.h"
#include "database/database.h"
#include "song.h"

#include <qapplication.h>




// private helpers ------------------------------------------------------------
namespace 
{
  template<class T, class T2>
  void assignAndCheckForModification( T &value, const T2 &rvalue, bool &modified )
  { if ( value != rvalue )
    {
      value = rvalue;
      modified = true;
    }
  }

  void sanitizeUnicode( QString &victim )
  {
    for ( unsigned i = 0; i < victim.length(); i++ )
      if ( !victim[i].isPrint() )
	victim[i] = '?';
  }




  QString genreIdToString( int genre )
  {
    switch ( genre )
    {
      case 0:
	return "Blues";
      case 1:
	return "Classic Rock";
      case 2:
	return "Country";
      case 3:
	return "Dance";
      case 4:
	return "Disco";
      case 5:
	return "Funk";
      case 6:
	return "Grunge";
      case 7:
	return "Hip-Hop";
      case 8:
	return "Jazz";
      case 9:
	return "Metal";
      case 10:
	return "New Age";
      case 11:
	return "Oldies";
      case 12:
	return "Other";
      case 13:
	return "Pop";
      case 14:
	return "R&B";
      case 15:
	return "Rap";
      case 16:
	return "Reggae";
      case 17:
	return "Rock";
      case 18:
	return "Techno";
      case 19:
	return "Industrial";
      case 20:
	return "Alternative";
      case 21:
	return "Ska";
      case 22:
	return "Death Metal";
      case 23:
	return "Pranks";
      case 24:
	return "Soundtrack";
      case 25:
	return "Euro-Techno";
      case 26:
	return "Ambient";
      case 27:
	return "Trip-Hop";
      case 28:
	return "Vocal";
      case 29:
	return "Jazz+Funk";
      case 30:
	return "Fusion";
      case 31:
	return "Trance";
      case 32:
	return "Classical";
      case 33:
	return "Instrumental";
      case 34:
	return "Acid";
      case 35:
	return "House";
      case 36:
	return "Game";
      case 37:
	return "Sound Clip";
      case 38:
	return "Gospel";
      case 39:
	return "Noise";
      case 40:
	return "AlternRock";
      case 41:
	return "Bass";
      case 42:
	return "Soul";
      case 43:
	return "Punk";
      case 44:
	return "Space";
      case 45:
	return "Meditative";
      case 46:
	return "Instrumental Pop";
      case 47:
	return "Instrumental Rock";
      case 48:
	return "Ethnic";
      case 49:
	return "Gothic";
      case 50:
	return "Darkwave";
      case 51:
	return "Techno-Industrial";
      case 52:
	return "Electronic";
      case 53:
	return "Pop-Folk";
      case 54:
	return "Eurodance";
      case 55:
	return "Dream";
      case 56:
	return "Southern Rock";
      case 57:
	return "Comedy";
      case 58:
	return "Cult";
      case 59:
	return "Gangsta";
      case 60:
	return "Top 40";
      case 61:
	return "Christian Rap";
      case 62:
	return "Pop/Funk";
      case 63:
	return "Jungle";
      case 64:
	return "Native American";
      case 65:
	return "Cabaret";
      case 66:
	return "New Wave";
      case 67:
	return "Psychadelic";
      case 68:
	return "Rave";
      case 69:
	return "Showtunes";
      case 70:
	return "Trailer";
      case 71:
	return "Lo-Fi";
      case 72:
	return "Tribal";
      case 73:
	return "Acid Punk";
      case 74:
	return "Acid Jazz";
      case 75:
	return "Polka";
      case 76:
	return "Retro";
      case 77:
	return "Musical";
      case 78:
	return "Rock & Roll";
      case 79:
	return "Hard Rock";
      case 80:
	return "Folk";
      case 81:
	return "Folk-Rock";
      case 82:
	return "National Folk";
      case 83:
	return "Swing";
      case 84:
	return "Fast Fusion";
      case 85:
	return "Bebob";
      case 86:
	return "Latin";
      case 87:
	return "Revival";
      case 88:
	return "Celtic";
      case 89:
	return "Bluegrass";
      case 90:
	return "Avantgarde";
      case 91:
	return "Gothic Rock";
      case 92:
	return "Progressive Rock";
      case 93:
	return "Psychedelic Rock";
      case 94:
	return "Symphonic Rock";
      case 95:
	return "Slow Rock";
      case 96:
	return "Big Band";
      case 97:
	return "Chorus";
      case 98:
	return "Easy Listening";
      case 99:
	return "Acoustic";
      case 100:
	return "Humour";
      case 101:
	return "Speech";
      case 102:
	return "Chanson";
      case 103:
	return "Opera";
      case 104:
	return "Chamber Music";
      case 105:
	return "Sonata";
      case 106:
	return "Symphony";
      case 107:
	return "Booty Bass";
      case 108:
	return "Primus";
      case 109:
	return "Porn Groove";
      case 110:
	return "Satire";
      case 111:
	return "Slow Jam";
      case 112:
	return "Club";
      case 113:
	return "Tango";
      case 114:
	return "Samba";
      case 115:
	return "Folklore";
      case 116:
	return "Ballad";
      case 117:
	return "Power Ballad";
      case 118:
	return "Rhythmic Soul";
      case 119:
	return "Freestyle";
      case 120:
	return "Duet";
      case 121:
	return "Punk Rock";
      case 122:
	return "Drum Solo";
      case 123:
	return "A capella";
      case 124:
	return "Euro-House";
      case 125:
	return "Dance Hall";
      default:
	return "undefined";
    }
  }
}




// public helpers -------------------------------------------------------------
QString getFieldName( tSongField field )
{
  switch ( field )
  {
    case FIELD_ARTIST: return qApp->translate( "FieldDescriptions", "Artist" );
    case FIELD_PERFORMER: return qApp->translate( "FieldDescriptions", "Performer" );
    case FIELD_TITLE: return qApp->translate( "FieldDescriptions", "Title" );
    case FIELD_ALBUM: return qApp->translate( "FieldDescriptions", "Album" );
    case FIELD_TRACKNUMBER: return qApp->translate( "FieldDescriptions", "Track Number" );
    case FIELD_DURATION: return qApp->translate( "FieldDescriptions", "Duration" );
    case FIELD_GENRE: return qApp->translate( "FieldDescriptions", "Genre" );
    case FIELD_FULLPLAYCOUNT: return qApp->translate( "FieldDescriptions", "Play Count (full)" );
    case FIELD_PARTIALPLAYCOUNT: return qApp->translate( "FieldDescriptions", "Play Count (partial)" );
    case FIELD_PLAYCOUNT: return qApp->translate( "FieldDescriptions", "Play Count" );
    case FIELD_LASTPLAYED: return qApp->translate( "FieldDescriptions", "Last Played" );
    case FIELD_RATING: return qApp->translate( "FieldDescriptions", "Rating" );
    case FIELD_YEAR: return qApp->translate( "FieldDescriptions", "Year" );
    case FIELD_FILE: return qApp->translate( "FieldDescriptions", "File" );
    case FIELD_PATH: return qApp->translate( "FieldDescriptions", "Path" );
    case FIELD_SIZE: return qApp->translate( "FieldDescriptions", "File Size" );
    case FIELD_EXISTS_SINCE: return qApp->translate( "FieldDescriptions", "Exists since" );
    case FIELD_EXTRA_CRITERION_MATCH: return qApp->translate( "FieldDescriptions", "Match" );
    default: return qApp->translate( "FieldDescriptions", "<unknown>" );
  }
}





// tSong ----------------------------------------------------------------------
tSong::tSong()
    : UniqueId( SONG_UID_INVALID ),
    Duration( 0 ),
    FileSize( 0 ),
    ExistsSince( time( NULL ) ),
    LastModified( time( NULL ) ),
    FullPlayCount( 0 ),
    PartialPlayCount( 0 ),
    Rating( -1 ),
    CacheValid( false ),
    SongCollection( NULL )
{
}




tSong::~tSong()
{
}




tFilename tSong::filename()
{
  return Filename;
}




tFilename tSong::pathname()
{
  return Pathname;
}




tFilename tSong::filenameOnly()
{
  return FilenameOnly;
}




tUniqueId tSong::uniqueId()
{
  return UniqueId;
}




QString tSong::key()
{
  ensureInfoIsThere();
  return Key;
}




QString tSong::album()
{
  ensureInfoIsThere();
  return Album;
}




QString tSong::artist()
{
  ensureInfoIsThere();
  return Artist;
}




QString tSong::performer()
{
  ensureInfoIsThere();
  return Performer;
}




QString tSong::title()
{
  ensureInfoIsThere();
  return Title;
}




QString tSong::year()
{
  ensureInfoIsThere();
  return Year;
}




QString tSong::genre()
{
  ensureInfoIsThere();
  return Genre;
}




QString tSong::trackNumber()
{
  ensureInfoIsThere();
  return TrackNumber;
}




float tSong::duration()
{
  ensureInfoIsThere();
  return Duration;
}




time_t tSong::existsSince()
{
  return ExistsSince;
}




time_t tSong::lastModified()
{
  ensureInfoIsThere();
  return LastModified;
}




time_t tSong::lastPlayed()
{
  return LastPlayed;
}




int tSong::playCount()
{
  return FullPlayCount+PartialPlayCount;
}




int tSong::fullPlayCount()
{
  return FullPlayCount;
}




int tSong::partialPlayCount()
{
  return PartialPlayCount;
}




int tSong::rating()
{
  return Rating;
}




int tSong::fileSize()
{
  ensureInfoIsThere();
  return FileSize;
}




QDomNode tSong::serialize( QDomDocument &doc )
{
  ensureInfoIsThere();

  QDomElement result = doc.createElement( "song" );
  result.setAttribute( "filename", Filename );
  result.setAttribute( "uniqueid", UniqueId );
  result.setAttribute( "album", Album );
  result.setAttribute( "artist", Artist );
  result.setAttribute( "performer", Performer );
  result.setAttribute( "title", Title );
  result.setAttribute( "year", Year );
  result.setAttribute( "genre", Genre );
  result.setAttribute( "tracknumber", TrackNumber );
  result.setAttribute( "duration", Duration );

  result.setAttribute( "existssince", (int) ExistsSince );
  result.setAttribute( "lastplayed", (int) LastPlayed );
  result.setAttribute( "lastmodified", (int) LastModified );
  result.setAttribute( "fullplaycount", FullPlayCount );
  result.setAttribute( "partialplaycount", PartialPlayCount );
  result.setAttribute( "rating", Rating );
  result.setAttribute( "filesize", (int) FileSize );
  return result;
}




void tSong::deserialize( const char **attributes )
{
  QString fn;
  fn = lookupAttribute( "filename", attributes );
  if ( fn != Filename )
  {
    runtimeError( qApp->translate( "ErrorMessages", "deserializing to different filename" ) );
    return ;
  }

  UniqueId = lookupAttribute( "uniqueid", attributes ).toUInt();
  Album = lookupAttribute( "album", attributes );
  Artist = lookupAttribute( "artist", attributes );
  try
  {
    Performer = lookupAttribute( "performer", attributes );
  }
  catch (...)
  {
    Performer = qApp->translate( "ErrorMessages", "<Unkown, reread tags>" );
  }

  Title = lookupAttribute( "title", attributes );
  Year = lookupAttribute( "year", attributes );
  Genre = lookupAttribute( "genre", attributes );
  TrackNumber = lookupAttribute( "tracknumber", attributes );
  Duration = lookupAttribute( "duration", attributes ).toFloat();

  if ( hasAttribute( "existssince", attributes ) )
    ExistsSince = lookupAttribute( "existssince", attributes ).toUInt();
  else
    ExistsSince = time( NULL );

  if ( hasAttribute( "fullplaycount", attributes ) )
  {
    LastPlayed = lookupAttribute( "lastplayed", attributes ).toUInt();
    FullPlayCount = lookupAttribute( "fullplaycount", attributes ).toUInt();
    PartialPlayCount = lookupAttribute( "partialplaycount", attributes ).toUInt();
  }
  else
  {
    FullPlayCount = 0;
    PartialPlayCount = 0;
  }

  if ( hasAttribute( "lastmodified", attributes ) )
    LastModified = lookupAttribute( "lastmodified", attributes ).toUInt();
  else
  {
    // we'll believe that our mtime is way back, so we'll
    // update our tags.
    LastModified = 0;
  }

  if ( hasAttribute( "rating", attributes ) )
    Rating = lookupAttribute( "rating", attributes ).toInt();
  else
    Rating = -1;

  if ( hasAttribute( "filesize", attributes ) )
    FileSize = lookupAttribute( "filesize", attributes ).toInt();
  else
  {
    QFileInfo info( Filename );
    FileSize = info.size();
  }

  CacheValid = true;

  updateDependentInfo();
}




void tSong::setUniqueId( tUniqueId id )
{
  UniqueId = id;
}




void tSong::setFilename( tFilename const &filename )
{
  if ( SongCollection )
    SongCollection->noticeSongFilenameAboutToChange( this, Filename, filename );

  Filename = filename;
  QFileInfo info( filename );
  FilenameOnly = info.fileName();
  Pathname = info.dirPath( true );
  CacheValid = false;

  if ( SongCollection )
    SongCollection->noticeSongModified( this );
}




void tSong::invalidateCache()
{
  CacheValid = false;

  // Cache invalidation does not entail song modification, so
  // don't call noticeSongModified here.
}




namespace 
{
  template<class T>
  int easyCompare( const T &val1, const T &val2 )
  {
    if ( val1 < val2 ) return -1;
    if ( val2 < val1 ) return 1;
    return 0;
  }
}




int tSong::compare( tSong *other, tSongField field )
{
  QString me_val,other_val;
  switch ( field )
  {
    case FIELD_ARTIST:
      me_val = artist();
      other_val = other->artist();
      break;
    case FIELD_PERFORMER:
      me_val = performer();
      other_val = other->performer();
      break;
    case FIELD_TITLE:
      me_val = title();
      other_val = other->title();
      break;
    case FIELD_ALBUM:
      me_val = album();
      other_val = other->album();
      break;
    case FIELD_TRACKNUMBER:
      {
	QRegExp extract_numbers( "^\\s*([0-9]+)" );

	if ( extract_numbers.search( trackNumber() ) != -1 )
	{
	  int my_track = extract_numbers.cap( 1 ).toInt();
	  if ( extract_numbers.search( other->trackNumber() ) != -1 )
	  {
	    // both have track numbers, compare them
	    int other_track = extract_numbers.cap( 1 ).toInt();
	    return easyCompare( my_track, other_track );
	  }
	  else
	  {
	    // other is smaller since it has no track number
	    return 1;
	  }
	}
	else
	{
	  if ( extract_numbers.search( other->trackNumber() ) != -1 )
	    return -1; // "this" is smaller since it has no track number
	  else
	    return 0; // neither has a track number
	}
      }
    case FIELD_DURATION:
      return easyCompare( duration(), other->duration() );
    case FIELD_GENRE:
      me_val = genre();
      other_val = other->genre();
      break;

    case FIELD_FULLPLAYCOUNT:
      return easyCompare( fullPlayCount(), other->fullPlayCount() );
    case FIELD_PARTIALPLAYCOUNT:
      return easyCompare( partialPlayCount(), other->partialPlayCount() );
    case FIELD_PLAYCOUNT:
      return easyCompare( playCount(), other->playCount() );
    case FIELD_LASTPLAYED:
      return easyCompare( lastPlayed(), other->lastPlayed() );
    case FIELD_RATING:
      return easyCompare( rating(), other->rating() );

    case FIELD_YEAR:
      return easyCompare( year().toInt(), other->year().toInt() );
    case FIELD_FILE:
      me_val = filenameOnly();
      other_val = other->filenameOnly();
      break;
    case FIELD_PATH:
      me_val = pathname();
      other_val = other->pathname();
      break;
    case FIELD_SIZE:
      return easyCompare( fileSize(), other->fileSize() );
      break;
    default:
      return 0;
  }

  me_val = me_val.lower();
  other_val = other_val.lower();

  return me_val.localeAwareCompare( other_val );
}




void tSong::stripTag()
{
  stripTagInternal();
  readInfo();
}




void tSong::rewriteTag()
{
  QString my_album = album();
  QString my_artist = artist();
  QString my_performer = performer();
  QString my_title = title();
  QString my_year = year();
  QString my_genre = genre();
  QString my_track_number = trackNumber();

  stripTagInternal();

  setAlbum( my_album );
  setArtist( my_artist );
  setPerformer( my_performer );
  setTitle( my_title );
  setYear( my_year );
  setGenre( my_genre );
  setTrackNumber( my_track_number );
}




void tSong::ensureInfoIsThere()
{
  if ( !CacheValid )
    readInfo();
}





void tSong::readInfo()
{
  QFileInfo file_info( Filename );
  FileSize = file_info.size();
  LastModified = file_info.lastModified().toTime_t();

  updateDependentInfo();
}




void tSong::played( time_t when, bool full )
{
  if ( full )
    FullPlayCount++;
  else
    PartialPlayCount++;

  LastPlayed = when;
  if ( SongCollection )
    SongCollection->noticeSongModified( this );
}




void tSong::rate( int rating )
{
  Rating = rating;
  if ( SongCollection )
    SongCollection->noticeSongModified( this );
}




void tSong::resetStatistics()
{
  FullPlayCount = 0;
  PartialPlayCount = 0;
  if ( SongCollection )
    SongCollection->noticeSongModified( this );
}




void tSong::setCollection( tSongCollection *collection )
{
  SongCollection = collection;
}




void tSong::updateDependentInfo()
{
  Key = ( artist() + " " + performer() + " " + title() +
    " " + album() + " " + filename() ).lower().simplifyWhiteSpace();
}




// tMP3Song -------------------------------------------------------------------
class tMP3Song : public tSong
{
  public:
    QString mimeType()
    { return "audio/mpeg"; }

    void readInfo();

    void setAlbum( QString const &value );
    void setArtist( QString const &value );
    void setPerformer( QString const &value );
    void setTitle( QString const &value );
    void setYear( QString const &value );
    void setGenre( QString const &value );
    void setTrackNumber( QString const &value );

    void stripTagInternal();
};




static QString getID3FieldString( ID3_Tag &tag, ID3_FrameID id )
{
  ID3_Frame * frame( tag.Find( id ) );
  if ( frame )
  {
    if ( frame->Contains( ID3FN_TEXT ) )
    {
      QString result;

      ID3_Field &field( frame->Field( ID3FN_TEXT ) );
      ID3_TextEnc encoding = field.GetEncoding();
      switch ( encoding )
      {
	case ID3TE_NONE: 
	  return QString::null;
	case ID3TE_ISO8859_1: 
	  result = QString::fromLatin1( field.GetRawText() );
	  sanitizeUnicode( result );
	  return result;
	case ID3TE_UTF16: 
	case ID3TE_UTF16BE: 
	  result = QString::fromUcs2( field.GetRawUnicodeText() );
	  sanitizeUnicode( result );
	  return result;
	case ID3TE_UTF8: 
	  result = QString::fromUtf8( field.GetRawText() );
	  sanitizeUnicode( result );
	  return result;
	case ID3TE_NUMENCODINGS: 
	  qDebug( "ID3TE_NUMENCODINGS encountered while reading ID3 tag\n" );
	  return QString::null;
	default:
	  qDebug( "ID3TE_XXX constant not known during ID3 tag read\n" );
	  return QString::null;
      }
    }
    else
      return "";
  }
  else
    return "";
}




static void setID3FieldString( ID3_Tag &tag, ID3_FrameID id, QString const &value )
{
  ID3_Frame *frame = NULL;
 
  while ( ( frame = tag.Find( id ) ) )
  {
    frame = tag.RemoveFrame(frame);
    delete frame;
  }
 
  frame = new ID3_Frame( id );

  ID3_Field &field( frame->Field( ID3FN_TEXT ) );
  // FIXME UTF8 tagging doesn't work
  // field.SetEncoding( ID3TE_UTF8 );
  // field.Set( value.utf8() );
  field.Set( value.latin1() );

  tag.AttachFrame( frame );

  tag.Update();
}




void tMP3Song::readInfo()
{
  if ( Filename == QString::null )
    return ;
  ID3_Tag tag( Filename.latin1() );

  CacheValid = true;

  bool modified = false;
  assignAndCheckForModification( Album, getID3FieldString( tag, ID3FID_ALBUM ), modified );
  assignAndCheckForModification( Artist, getID3FieldString( tag, ID3FID_LEADARTIST ), modified );
  assignAndCheckForModification( Performer, getID3FieldString( tag, ID3FID_MIXARTIST ), modified );
  assignAndCheckForModification( Title, getID3FieldString( tag, ID3FID_TITLE ), modified );
  assignAndCheckForModification( Year, getID3FieldString( tag, ID3FID_YEAR ), modified );
  QString genre = getID3FieldString( tag, ID3FID_CONTENTTYPE );
  assignAndCheckForModification( TrackNumber, getID3FieldString( tag, ID3FID_TRACKNUM ), modified );

  mp3info info;
  info.filename = const_cast< char * >( Filename.latin1() );
  info.file = fopen( info.filename, "r" );
  if ( info.file )
  {
    if ( get_mp3_info( &info, SCAN_QUICK, 0 ) && info.seconds > 0 )
      assignAndCheckForModification( Duration, 0, modified );
    else
      assignAndCheckForModification( Duration, info.seconds, modified );
    fclose( info.file );
  }
  else
    assignAndCheckForModification( Duration, 0, modified );

  QRegExp re( "^\\(([0-9]+)\\)$" );
  if ( re.exactMatch( genre )  )
    assignAndCheckForModification( Genre, 
      genreIdToString( atoi( re.cap( 1 ).latin1() ) ), modified );
  else
    assignAndCheckForModification( Genre, genre, modified );

  if ( modified )
  {
    if ( SongCollection )
      SongCollection->noticeSongModified( this );
  }

  tSong::readInfo();
}




void tMP3Song::setAlbum( QString const &value )
{
  ID3_Tag tag( Filename.latin1() );
  setID3FieldString( tag, ID3FID_ALBUM, value );
  readInfo();
}




void tMP3Song::setArtist( QString const &value )
{
  ID3_Tag tag( Filename.latin1() );
  setID3FieldString( tag, ID3FID_LEADARTIST, value );
  readInfo();
}




void tMP3Song::setPerformer( QString const &value )
{
  ID3_Tag tag( Filename.latin1() );
  setID3FieldString( tag, ID3FID_MIXARTIST, value );
  readInfo();
}




void tMP3Song::setTitle( QString const &value )
{
  ID3_Tag tag( Filename.latin1() );
  setID3FieldString( tag, ID3FID_TITLE, value );
  readInfo();
}




void tMP3Song::setYear( QString const &value )
{
  ID3_Tag tag( Filename.latin1() );
  setID3FieldString( tag, ID3FID_YEAR, value );
  readInfo();
}




void tMP3Song::setGenre( QString const &value )
{
  ID3_Tag tag( Filename.latin1() );
  setID3FieldString( tag, ID3FID_CONTENTTYPE, value );
  readInfo();
}




void tMP3Song::setTrackNumber( QString const &value )
{
  ID3_Tag tag( Filename.latin1() );
  setID3FieldString( tag, ID3FID_TRACKNUM, value );
  readInfo();
}




void tMP3Song::stripTagInternal()
{
  ID3_Tag tag( Filename.latin1() );
  tag.Strip();
}




// tOggSong -------------------------------------------------------------------
class tOggSong : public tSong
{
  public:
    QString mimeType()
    { return "audio/x-ogg"; }

    void readInfo();

    void setAlbum( QString const &value );
    void setArtist( QString const &value );
    void setPerformer( QString const &value );
    void setTitle( QString const &value );
    void setYear( QString const &value );
    void setGenre( QString const &value );
    void setTrackNumber( QString const &value );

    void stripTagInternal();
};




static QString getOggFieldString( vorbis_comment *vc, const char *tag )
{
  int count = vorbis_comment_query_count( vc, const_cast<char *>(tag) );
  if ( count > 0 )
  {
    QString result = QString::fromUtf8( vorbis_comment_query( vc, const_cast<char *>(tag), 0 ) );
    sanitizeUnicode( result );
    return result;
  }
  else
    return "";
}




// ----------------------------------------------------------------------------
// begin code shamelessly pulled from xmms, with minimal changes
// ----------------------------------------------------------------------------
static char** get_comment_list(vorbis_comment *vc)
{
  int i;
  char **strv;

  strv = g_new0(char*, vc->comments + 1);
  for (i = 0; i < vc->comments; i++)
  {
    strv[i] = g_strdup(vc->user_comments[i]);
  }

  return strv;
}

static char** add_tag(char **list, const char *label, const char *in_tag)
{
  char **ptr = list, *reallabel = g_strconcat(label, "=", NULL);

  char *tag = g_strdup( in_tag );
  g_strstrip(tag);
  if (strlen(tag) == 0)
    tag = NULL;
  /*
   * There can be several tags with the same label.  We clear
   * them all.
   */
  while (*ptr != NULL)
  {
    if (!g_strncasecmp(reallabel, *ptr, strlen(reallabel)))
    {
      g_free(*ptr);
      if (tag != NULL)
      {
	*ptr = g_strconcat(reallabel, tag, NULL);
	g_free(tag);
	tag = NULL;
	ptr++;
      }
      else
      {
	char **str;
	for (str = ptr; *str; str++)
	  *str = *(str + 1);
      }

    }
    else
      ptr++;
  }
  if (tag)
  {
    int i = 0;
    for (ptr = list; *ptr; ptr++)
      i++;
    list = g_renew(char*, list, i + 2);
    list[i] = g_strconcat(reallabel, tag, NULL);
    list[i + 1] = NULL;
    g_free(tag);
  }
  g_free(reallabel);

  return list;
}

static void add_list(vorbis_comment *vc, char **comments)
{
  while (*comments)
    vorbis_comment_add(vc, *comments++);
}
// ----------------------------------------------------------------------------
// end code shamelessly pulled from xmms
// ----------------------------------------------------------------------------




static void setOggFieldString( const QString &filename, QString const &key, QString const &value )
{
  vcedit_state *state = vcedit_new_state();

  FILE *ogg_song = fopen( filename.latin1(), "rb" );
  if ( ogg_song == NULL )
    runtimeError( qApp->translate( "ErrorMessages", "Trouble reading from %1" ).arg( filename ) );

  if ( vcedit_open( state, ogg_song ) )
  {
    fclose( ogg_song );
    runtimeError( qApp->translate( "ErrorMessages", "Trouble reading tags from %1" ).arg( filename ) );
  }
  
  vorbis_comment *vc = vcedit_comments( state );
  char **comment_list = get_comment_list( vc );
  comment_list = add_tag( comment_list, key.latin1(), value.latin1() );
  vorbis_comment_clear( vc );
  add_list( vc, comment_list );

  QString new_ogg_filename = filename + ".madman_retag";
  FILE *new_ogg_song = fopen( new_ogg_filename.latin1(), "wb" );
  if ( new_ogg_song == NULL )
  {
    fclose( ogg_song );
    runtimeError( qApp->translate( "ErrorMessages", "Trouble writing retagged file" ) );
  }

  if ( vcedit_write( state, new_ogg_song ) )
  {
    fclose( ogg_song );
    fclose( new_ogg_song );
    remove( new_ogg_filename.latin1() );
    runtimeError( qApp->translate( "ErrorMessages", "Trouble writing retagged file" ) );
  }

  fclose( new_ogg_song );
  fclose( ogg_song );
  if ( rename( new_ogg_filename.latin1(), filename.latin1() ) )
    remove( new_ogg_filename.latin1() );

  vcedit_clear( state );
}







void tOggSong::readInfo()
{
  if ( Filename == QString::null )
    return ;
  FILE *ogg_song = fopen( Filename.latin1(), "rb" );
  if ( ogg_song == NULL )
  {
    runtimeError( qApp->translate( "ErrorMessages", "Trouble reading tags from %1" ).arg( Filename ) );
  }
  OggVorbis_File vf;
  if ( ov_open( ogg_song, &vf, NULL, 0 ) )
  {
    fclose( ogg_song );
    runtimeError( qApp->translate( "ErrorMessages", "Trouble reading tags from %1" ).arg( Filename ) );
  }

  // ov_open has taken possession of ogg_song, so no need to fclose here.

  vorbis_comment *vc = ov_comment( &vf, -1 );

  CacheValid = true;

  bool modified = false;

  assignAndCheckForModification( Album, getOggFieldString( vc, "ALBUM" ), modified );
  assignAndCheckForModification( Artist, getOggFieldString( vc, "ARTIST" ), modified );
  assignAndCheckForModification( Performer, getOggFieldString( vc, "PERFORMER" ), modified );
  assignAndCheckForModification( Title, getOggFieldString( vc, "TITLE" ), modified );
  assignAndCheckForModification( Year, getOggFieldString( vc, "YEAR" ), modified );
  assignAndCheckForModification( Genre, getOggFieldString( vc, "GENRE" ), modified );
  assignAndCheckForModification( TrackNumber, getOggFieldString( vc, "TRACKNUMBER" ), modified );
  assignAndCheckForModification( Duration, ov_time_total( &vf, -1 ), modified );

  ov_clear( &vf );

  if ( modified )
  {
    if ( SongCollection )
      SongCollection->noticeSongModified( this );
  }

  tSong::readInfo();
}




void tOggSong::setAlbum( QString const &value )
{
  setOggFieldString( Filename, "ALBUM", value );
  readInfo();
}




void tOggSong::setArtist( QString const &value )
{
  setOggFieldString( Filename, "ARTIST", value );
  readInfo();
}




void tOggSong::setPerformer( QString const &value )
{
  setOggFieldString( Filename, "PERFORMER", value );
  readInfo();
}




void tOggSong::setTitle( QString const &value )
{
  setOggFieldString( Filename, "TITLE", value );
  readInfo();
}




void tOggSong::setYear( QString const &value )
{
  setOggFieldString( Filename, "YEAR", value );
  readInfo();
}




void tOggSong::setGenre( QString const &value )
{
  setOggFieldString( Filename, "GENRE", value );
  readInfo();
}




void tOggSong::setTrackNumber( QString const &value )
{
  setOggFieldString( Filename, "TRACKNUMBER", value );
  readInfo();
}




void tOggSong::stripTagInternal()
{
  // *** FIXME: NYI
}




// public ---------------------------------------------------------------------
tSong *makeSong( tFilename filename )
{
  QRegExp mp3_pattern( "\\.mp3$", false );
  mp3_pattern.setCaseSensitive( false );
  QRegExp ogg_pattern( "\\.ogg$", false );
  ogg_pattern.setCaseSensitive( false );

  if ( mp3_pattern.search( filename ) >= 0 )
  {
    tSong *song = new tMP3Song;
    song->setFilename( filename );
    return song;
  }
  if ( ogg_pattern.search( filename ) >= 0 )
  {
    tSong *song = new tOggSong;
    song->setFilename( filename );
    return song;
  }
  runtimeError( qApp->translate( "ErrorMessages", "Unknown file type: %1" ).arg(  filename ) );
  throw 5; // bogus, to silence compiler warnings
}





tSong *deserializeSong( const char **attributes)
{
  QString filename;
  filename = lookupAttribute( "filename", attributes );

  tSong *song = makeSong( filename );
  song->deserialize( attributes );
  return song;
}
