/*
madman - a music manager
Copyright (C) 2003  Andreas Kloeckner <ak@ixion.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/




#ifndef SONG
#define SONG




#include <qstring.h>
#include <qdom.h>
#include <qobject.h>
#include "utility/base.h"




#define SONG_UID_INVALID (-1)




enum tSongField {
  FIELD_ARTIST = 0,
  FIELD_PERFORMER = 1,
  FIELD_TITLE = 2,
  FIELD_ALBUM = 3,
  FIELD_TRACKNUMBER = 4,
  FIELD_DURATION = 5,
  FIELD_GENRE = 6,

  FIELD_FULLPLAYCOUNT = 7,
  FIELD_PARTIALPLAYCOUNT = 8,
  FIELD_PLAYCOUNT = 9,
  FIELD_LASTPLAYED = 10,
  FIELD_RATING = 11,

  FIELD_YEAR = 12,
  FIELD_FILE = 13,
  FIELD_PATH = 14,
  FIELD_SIZE = 15,

  FIELD_EXISTS_SINCE = 16,
  FIELD_REAL_COUNT = 17,
  
  FIELD_EXTRA_FIRST = ( FIELD_REAL_COUNT),
  FIELD_EXTRA_CRITERION_MATCH = ( FIELD_EXTRA_FIRST + 0 ),
  FIELD_COUNT = ( FIELD_REAL_COUNT + 1)
};




QString getFieldName( tSongField field );




class tSongCollection;




class tSong 
{
  protected:
    tFilename Filename, Pathname, FilenameOnly;
    tUniqueId UniqueId;

    QString Album, Artist, Performer, Title, Year, Genre, TrackNumber;
    QString Key;
    float Duration;
    long int FileSize;

    time_t ExistsSince, LastPlayed, LastModified;
    int FullPlayCount,PartialPlayCount;
    // -1 if unspecified
    int Rating;
    
    bool CacheValid;

    tSongCollection *SongCollection;

  public:
    tSong();
    virtual ~tSong();

    tFilename filename();
    tFilename pathname();
    tFilename filenameOnly();
    tUniqueId uniqueId();

    /**
     * This function returns a lowercased concatenation of the
     * most characterizing metadata strings of the song:
     * artist, performer, title, album, fully qualified filename.
     */
    QString key();

    QString album();
    QString artist();
    QString performer();
    QString title();
    QString year();
    QString genre();
    QString trackNumber();
    float duration();

    time_t existsSince();
    time_t lastModified();
    time_t lastPlayed();
    int playCount();
    int fullPlayCount();
    int partialPlayCount();
    int rating();
    
    int fileSize();
    virtual QString mimeType() = 0;

    virtual QDomNode serialize( QDomDocument &doc );
    virtual void deserialize( const char **attributes );

    void setUniqueId( tUniqueId id );
    void setFilename( tFilename const &filename );
    void invalidateCache();

    int compare( tSong *other, tSongField field );

    virtual void setAlbum( QString const &value ) = 0;
    virtual void setArtist( QString const &value ) = 0;
    virtual void setPerformer( QString const &value ) = 0;
    virtual void setTitle( QString const &value ) = 0;
    virtual void setYear( QString const &value ) = 0;
    virtual void setGenre( QString const &value ) = 0;
    virtual void setTrackNumber( QString const &value ) = 0;

    void stripTag();
    void rewriteTag();

    void ensureInfoIsThere();
    /* tSong::readInfo needs to be called *after* any information
     * has been read by the subclass.
     */
    virtual void readInfo();

    void played( time_t when, bool full );
    void rate( int rating );
    void resetStatistics();

    void setCollection( tSongCollection *collection );

  protected:
    virtual void stripTagInternal() = 0;

    /** Updates dependent information that is too easy to come
     * by to cache on disk, but too hard to compute on the fly.
     */
    void updateDependentInfo();
};




class less_song
{
  private:
    tSongField		Field;
    bool		Ascending;

  public:
    less_song( tSongField field, bool ascending = true )
      : Field( field ), Ascending( ascending )
      { }
    bool operator()( tSong *song1, tSong *song2 )
    {
      if ( Ascending)
	return song1->compare( song2, Field ) < 0;
      else
	return -song1->compare( song2, Field ) < 0;
    }
};




tSong *makeSong( tFilename filename );
tSong *deserializeSong( const char **attributes );




#endif 
