/*
madman - a music manager
Copyright (C) 2003  Andreas Kloeckner <ak@ixion.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/




#include "utility/progress.h"

#include <dirent.h>
#include <sys/stat.h>
#include <qapplication.h>
#include <qprogressdialog.h>
#include <qdir.h>

#include "song_collection.h"

#include "song_collection.moc"




tSongCollection::tSongCollection()
    : NextUniqueId( 0 )
{}




void tSongCollection::readTags( tProgress *progress,
    const QString &progresstext, 
    tSongList *list )
{
  if ( list == NULL ) list = &SongList;
  progress->setWhat( progresstext );
  progress->setTotalSteps( list->size() );
  int index = 0;

  FOREACH( first, *list, tSongList )
  {
    (*first)->ensureInfoIsThere();

    ++index;
    if ( index % 10 == 0 )
    {
      progress->setProgress( index );
      if ( progress->wasCancelled() )
	break;
    }
  }
}




void tSongCollection::rereadTags( tProgress *progress )
{
  progress->setWhat( tr( "Reading tags..." ) );
  progress->setTotalSteps( SongList.size() );
  int index = 0;

  FOREACH( first, SongList, tSongList )
  {
    (*first)->invalidateCache();
    (*first)->ensureInfoIsThere();

    ++index;
    if ( index % 10 == 0 )
    {
      progress->setProgress( index );
      if ( progress->wasCancelled() )
	break;
    }
  }
}




void tSongCollection::scan( tDirectoryList const &dl, tProgress *progress )
{
  emit notifyBeginBulkChange();

  // scan for new songs
  progress->setWhat( tr( "Scanning for files..." ) );

  New.clear();
  FOREACH_CONST( first, dl, tDirectoryList )
    scanSingleDirectory( *first, progress );

  // look for removed songs
  progress->setWhat( tr( "Looking for unwanted songs..." ) );
  progress->setTotalSteps( SongList.size() );
  int index = 0;

  tSongList unwanted_songs;
  FOREACH( first, SongList, tSongList )
  {
    bool found = false;
    FOREACH_CONST( first_d, dl, tDirectoryList )
      if ( ( *first ) ->filename().left( first_d->length() ) == *first_d )
	found = true;
    if ( !found )
      unwanted_songs.push_back( *first );

    ++index;
    if ( index % 10 == 0 )
    {
      progress->setProgress( index );
    }
  }

  // remove unwanted songs
  FOREACH( first, unwanted_songs, tSongList )
  {
    cerr << "unwanted: " << ( *first ) ->filename() << endl;
    remove ( *first );
  }

  // look for orphans
  progress->setWhat( tr( "Looking for orphans..." ) );
  progress->setTotalSteps( SongList.size() );
  index = 0;

  Orphans.clear();
  FOREACH( first, SongList, tSongList )
  {
    struct stat statbuf;
    if ( stat( ( *first ) ->filename().latin1(), &statbuf ) )
      Orphans.push_back( *first );

    ++index;
    if ( index % 10 == 0 )
    {
      progress->setProgress( index );
    }
  }

  // read tags now so move detection can use them.
  readTags( progress, tr( "Updating existing tags..." ) );
  readTags( progress, tr( "Reading tags of new songs..." ), &New );

  // compare them to new songs to detect moves
  progress->setWhat( tr( "Detecting moves..." ) );
  progress->setTotalSteps( Orphans.size() );
  index = 0;

  tSongList::iterator first_o = Orphans.begin(), last_o = Orphans.end();
  while ( first_o != last_o )
  {
    bool erase_orphan = false;
    tSongList::iterator first_n = New.begin(), last_n = New.end();
    while ( first_n != last_n )
    {
      if ( ( *first_n ) ->album() == ( *first_o ) ->album()
           && ( *first_n ) ->artist() == ( *first_o ) ->artist()
           && ( *first_n ) ->title() == ( *first_o ) ->title() )
      {
        cerr << "move: " << ( *first_o ) ->filename() << " -> " << ( *first_n ) ->filename() << endl;

        ( *first_o )->setFilename( ( *first_n ) ->filename() );
        first_n = New.erase( first_n );
        last_n = New.end();

        erase_orphan = true;
        break;
      }
      else
        first_n++;
    }
    if ( erase_orphan )
    {
      first_o = Orphans.erase( first_o );
      last_o = Orphans.end();
    }
    else
      first_o++;

    progress->setProgress( ++index );
  }

  // add new songs
  progress->setWhat( tr( "Adding songs..." ) );
  progress->setTotalSteps( New.size() );
  index = 0;

  FOREACH( first, New, tSongList )
  {
    cerr << "add: " << ( *first ) ->filename() << endl;
    insert( *first, false );

    ++index;
    if ( index % 10 == 0 )
    {
      progress->setProgress( index );
    }
  }

  // remove orphaned songs
  FOREACH( first, Orphans, tSongList )
  {
    cerr << "delete: " << ( *first ) ->filename() << endl;
    remove( *first );
  }

  // reread tags of songs that were modified since last scan
  progress->setWhat( tr( "Updating tags of changed songs..." ) );
  progress->setTotalSteps( SongList.size() );

  index = 0;
  FOREACH( first, SongList, tSongList )
  {
    QFileInfo file_info( (*first)->filename() );
    if ( (time_t) file_info.lastModified().toTime_t() > (*first)->lastModified() )
    {
      cerr << "changed: " << ( *first ) ->filename() << endl;
      (*first)->invalidateCache();
      (*first)->ensureInfoIsThere();
    }

    ++index;
    if ( index % 10 == 0 )
    {
      progress->setProgress( index );
    }
  }

  emit notifyEndBulkChange();
}




void tSongCollection::insert( tSong *song, bool loaded )
{
  if ( song->uniqueId() == SONG_UID_INVALID )
    song->setUniqueId( nextUniqueId() );
  SongList.push_back( song );
  FilenameMap.insert( make_pair( song->filename(), song ) );
  UniqueIdMap.insert( make_pair( song->uniqueId(), song ) );
  if ( loaded )
    emit notifyLoad( song );
  else
    emit notifyAddition( song );

  song->setCollection( this );
}




void tSongCollection::remove( tSong *song )
{
  emit notifyDeletion( song );
  FilenameMap.erase( song->filename() );
  UniqueIdMap.erase( song->uniqueId() );

  FOREACH( first, SongList, tSongList )
  if ( *first == song )
  {
    SongList.erase( first );
    break;
  }
}




void tSongCollection::clear()
{
  emit notifyBeginBulkChange();
  while ( SongList.size() > 0 )
    remove( SongList[ SongList.size() - 1 ] );
  emit notifyEndBulkChange();
}




void tSongCollection::noticeSongModified( tSong *song )
{
  emit notifySongModified( song );
}




void tSongCollection::noticeSongFilenameAboutToChange( tSong *song, const tFilename &old, const tFilename &current )
{
  FilenameMap.erase( old );
  FilenameMap.insert( make_pair( current, song ) );
}




void tSongCollection::beginBulkChange()
{
  emit notifyBeginBulkChange();
}




void tSongCollection::endBulkChange()
{
  emit notifyEndBulkChange();
}




QDomNode tSongCollection::serialize( QDomDocument &doc )
{
  QDomElement result = doc.createElement( "songs" );
  FOREACH( first, SongList, tSongList )
    result.appendChild( ( *first ) ->serialize( doc ) );
  return result;
}




tSong *tSongCollection::getByFilename( tFilename const &filename )
{
  tFilenameMap::iterator it = FilenameMap.find( filename );
  if ( it == FilenameMap.end() )
    return NULL;
  else
    return it->second;
}




tSong *tSongCollection::getByUniqueId( tUniqueId const &uid )
{
  tUniqueIdMap::iterator it = UniqueIdMap.find( uid );
  if ( it == UniqueIdMap.end() )
    return NULL;
  else
    return it->second;
}




bool tSongCollection::isValid( tSong *song )
{
  FOREACH( first, SongList, tSongList )
    if ( (*first) == song )
      return true;
  return false;
}




tUniqueId tSongCollection::nextUniqueId()
{
  while ( getByUniqueId( NextUniqueId ) )
    NextUniqueId++;
  return NextUniqueId++;
}




void tSongCollection::scanSingleDirectory( tFilename const &dir, tProgress *progress )
{
  tFilename realdir = dir;
  if ( realdir.endsWith( "/" ) )
    realdir = realdir.left( realdir.length() - 1 );

  QDir directory( realdir );
  if ( !directory.exists() )
  {
    cerr << "[ERROR] nonexistent: " << realdir.latin1() << endl;
    return ;
  }

  QStringList strlist = directory.entryList();
  for ( unsigned i = 0; i < strlist.size(); i++ )
  {
    progress->processEvents();

    tFilename name = strlist[ i ];
    tFilename fullname = realdir + "/" + name;

    QFileInfo file( fullname );
    if ( !file.exists() )
    {
      cerr << "[ERROR] stat failed: " << fullname << endl;
      continue;
    }
    if ( file.isDir() )
    {
      if ( name != "." && name != ".." )
        scanSingleDirectory( fullname, progress );
    }
    else
    {
      tSong *song = getByFilename( realdir + "/" + name );
      if ( !song )
      {
	try 
	{
	  song = makeSong( fullname );
          New.push_back( song );
	}
	catch (...)
	{
	}
      }
    }
  }
}
