/*
madman - a music manager
Copyright (C) 2003  Andreas Kloeckner <ak@ixion.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/




#ifndef _HEADER_SEEN_SONGSET_H
#define _HEADER_SEEN_SONGSET_H




#include <algorithm>
#include <memory>
#include "song_collection.h"




// criterion string helper ----------------------------------------------------
QString quoteString( QString victim );







// tRendering -----------------------------------------------------------------
class tProgress;
struct tRenderingEntry
{
  float	CriterionMatch;
  tSong	*Song;
};

typedef vector<tRenderingEntry> tRendering;

class less_rendering
{
  private:
    tSongField		Field;
    bool		Ascending;

  public:
    less_rendering( tSongField field, bool ascending = true )
      : Field( field ), Ascending( ascending )
      { }
    bool operator()( const tRenderingEntry &entry1, const tRenderingEntry &entry2 )
    {
      if ( Field != FIELD_EXTRA_CRITERION_MATCH )
      {
	if ( Ascending)
	  return entry1.Song->compare( entry2.Song, Field ) < 0;
	else
	  return -entry1.Song->compare( entry2.Song, Field ) < 0;
      }
      else
      {
	if ( Ascending )
	  return entry1.CriterionMatch < entry2.CriterionMatch;
	else
	  return entry1.CriterionMatch > entry2.CriterionMatch;
      }
    }
};




// tSongSet -------------------------------------------------------------------
class tProgress;
class tCriterion;

class tSongSet : public QObject
{
    Q_OBJECT

  protected:
    tSongCollection *SongCollection;

    QString Criterion;
    auto_ptr<tCriterion> ParsedCriterion;

    tRendering Rendering;
    tUniqueIdSet RenderingSet,PositiveSet,NegativeSet;

    int BulkChangeLevel;
    bool SubstantialBulkChange;
    
    float MatchThreshold;

  public:
    tSongSet();
    ~tSongSet()
    { }

    /** This subroutine first clears the content of this song
     * set, then switches over to the new collection, then
     * reevaluates the criterion.
     */
    void setSongCollection( tSongCollection *collection );

    tSongSet &operator=( const tSongSet &src );

    QDomNode serialize( QDomDocument &doc );
    void setRendering( const tUniqueIdList &list );
    void setPositiveSet( const tUniqueIdList &list );
    void setNegativeSet( const tUniqueIdList &list );

    void render( tSongList &rendering );
    void render( tRendering &rendering );
    void render( tUniqueIdList &rendering );

    bool containsSong( tUniqueId id );

    void clear();
    void sortBy( tSongField field, bool ascending = true );
    void jumble();
    void reverse();

    void add( tSong *song, int preferred_index = -1 );
    void add( const tFilename &filename, int preferred_index = -1 );
    void remove( int index );
    void remove( tSong *song );

    QString const &criterion() const
    {
      return Criterion;
    }
    /// will free the criterion when destroyed
    void setCriterion( tCriterion *crit );
    void setCriterion( const QString &crit, bool use_fuzzy = false );
    void reevaluateCriterion( tProgress *progess = NULL );
    bool hasBinaryCriterion();

    bool isInRendering( tUniqueId uid )
    {
      return RenderingSet.count( uid ) != 0;
    }

  signals:
    void notifyChange( bool substanial_change, tProgress *progress );
    void notifySongModified( tSong *song );

  public slots:
    void beginBulkChange();
    void endBulkChange();

  protected slots:
    void noticeCollectionAddition( tSong *song );
    void noticeCollectionDeletion( tSong *song );

    void noticeSongModified( tSong *tSong );

  protected:
    void hasChanged( bool substantial_change, tProgress *progress = NULL );

    void addToRendering( tSong *song, float match, int preferred_index = -1 );
    void removeFromRendering( tSong *song );
    void removeFromRendering( int index );

  private:
    // disabled
    tSongSet( const tSongSet &src );
};




// tCriterion -----------------------------------------------------------------
class tCriterion
{
  public:
    virtual ~tCriterion()
    { }
    virtual float operator() ( tSong *song ) const = 0;
    virtual bool isBinary() = 0;
};




class tInSetCriterion : public tCriterion
{
    tSongSet *SongSet;
  public:
    tInSetCriterion( tSongSet *ss )
    : SongSet( ss )
    { }
    float operator() ( tSong *song ) const
    {
      return SongSet->containsSong( song->uniqueId() );
    }
    bool isBinary()
    { return true; }
};




class tOrCriterion : public tCriterion
{
    auto_ptr<tCriterion>	Operand1,Operand2;
  public:
    tOrCriterion( tCriterion *op1, tCriterion *op2 )
      : Operand1( op1 ), Operand2( op2 )
    { }
    float operator() ( tSong *song ) const
    {
      return max( (*Operand1)( song ), (*Operand2)( song ) );
    }
    bool isBinary() 
    { 
      return Operand1->isBinary() && Operand2->isBinary(); 
    }
};




class tAndCriterion : public tCriterion
{
    auto_ptr<tCriterion>	Operand1,Operand2;
  public:
    tAndCriterion( tCriterion *op1, tCriterion *op2 )
      : Operand1( op1 ), Operand2( op2 )
    { }
    float operator() ( tSong *song ) const
    {
      return min( (*Operand1)( song ), (*Operand2)( song ) );
    }
    bool isBinary() 
    { 
      return Operand1->isBinary() && Operand2->isBinary(); 
    }
};




class tNotCriterion : public tCriterion
{
    auto_ptr<tCriterion>	Operand;
  public:
    tNotCriterion( tCriterion *op )
      : Operand( op )
    { }
    float operator() ( tSong *song ) const
    {
      return 1 - (*Operand)( song );
    }
    bool isBinary() 
    { 
      return Operand->isBinary(); 
    }
};




auto_ptr<tCriterion> parseCriterion( QString criterion, bool use_fuzzy );




#endif
