/*
madman - a music manager
Copyright (C) 2003  Andreas Kloeckner <ak@ixion.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/




#include "utility/player.h"
#include "database/database.h"
#include "ui/overview.h"
#include "ui/mainwin.h"
#include <qpopupmenu.h>
#include <qlistview.h>
#include <algorithm>

#include "overview.moc"




// private helpers ------------------------------------------------------------
namespace
{
  QString normalize( const QString &victim )
  {
    return victim.lower();
  }
}




// tOverviewItem --------------------------------------------------------------
tOverviewItem::tOverviewItem( QListView *view,  QListViewItem *after, const QString &label, const QString &criterion )
: QListViewItem( view, after, label ), Criterion( criterion )
{ 
}




tOverviewItem::tOverviewItem( QListViewItem *parent, QListViewItem *after, const QString &label, const QString &criterion )
: QListViewItem( parent, after, label ), Criterion( criterion )
{
}





// tExpandableOverviewItem --------------------------------------------------------------
tExpandableOverviewItem::tExpandableOverviewItem( QListView *view,  QListViewItem *after, const QString &label, const QString &criterion )
: tOverviewItem( view, after, label, criterion )
{ 
  setExpandable( true );
  HasExpanded = false;
}




tExpandableOverviewItem::tExpandableOverviewItem( QListViewItem *parent, QListViewItem *after, const QString &label, const QString &criterion )
: tOverviewItem( parent, after, label, criterion )
{
  setExpandable( true );
  HasExpanded = false;
}




void tExpandableOverviewItem::setOpen( bool o )
{
  if ( !HasExpanded && o )
  {
    listView()->setUpdatesEnabled( false );
    addExpansion();
    listView()->setUpdatesEnabled( true );
    HasExpanded = true;
  }
  tOverviewItem::setOpen( o );
}




void tExpandableOverviewItem::addExpansion()
{
}




// tArtistsOverviewItem -------------------------------------------------------
tArtistsOverviewItem::tArtistsOverviewItem( tDatabase &database, QListView *view,  QListViewItem *after, const QString &label, const QString &criterion )
: tExpandableOverviewItem( view, after, label, criterion ), 
  Database( database )
{ 
}




tArtistsOverviewItem::tArtistsOverviewItem( tDatabase &database, QListViewItem *parent, QListViewItem *after, const QString &label, const QString &criterion )
: tExpandableOverviewItem( parent, after, label, criterion ),
  Database( database )
{ 
}




void tArtistsOverviewItem::addExpansion()
{
  tSongList songs( Database.SongCollection.begin(), Database.SongCollection.end() );
  ::sort( songs.begin(), songs.end(), less_song( FIELD_ARTIST ) );
  QListViewItem *last_item = NULL;
  QString last_entry = QString::null;

  int artist_count = 0;

  tSongList::iterator first = songs.begin(), last = songs.end();
  while ( first != last )
  {
    QString this_nonnormalized_entry = (*first)->artist();
    QString this_entry = normalize( this_nonnormalized_entry );
    if ( this_nonnormalized_entry == "" )
      this_nonnormalized_entry = "<none>";

    tSongList this_songs;
    while ( first != last && this_entry == normalize( (*first)->artist() ) )
      this_songs.push_back( *first++ );

    ::sort( this_songs.begin(), this_songs.end(), less_song( FIELD_ALBUM ) );
    QListViewItem *entry_item = new tOverviewItem( this, last_item, this_nonnormalized_entry, QString( "~artist(complete:%1)" ).arg( quoteString( this_entry ) ) );

    int album_count = 0;
    int song_count = 0;
    int total_song_count = 0;

    QString last_subentry = QString::null;
    QListViewItem *last_subitem = NULL;

    FOREACH( sub_first, this_songs, tSongList )
    {
      QString this_nonnormalized_subentry = (*sub_first)->album();
      QString this_subentry = normalize( this_nonnormalized_subentry );
      if ( this_nonnormalized_subentry == "" )
	this_nonnormalized_subentry = "<none>";

      if ( this_subentry != last_subentry )
      {
	if ( last_subitem )
	{
	  last_subitem->setText( 0, last_subitem->text( 0 ) + QString(" (%1)").arg( song_count ) );
	  song_count = 0;
	}

	last_subitem = new tOverviewItem( entry_item, last_subitem, this_nonnormalized_subentry, 
	    QString( "~artist(complete:%1)&~album(complete:%2)" )
	    .arg( quoteString( this_entry ) )
	    .arg( quoteString( this_subentry ) ) );
	album_count++;
      }

      last_subentry = this_subentry;
      song_count++;
      total_song_count++;
    }
    if ( last_subitem )
      last_subitem->setText( 0, last_subitem->text( 0 ) + QString(" (%1)").arg( song_count ) );

    entry_item->setText( 0, this_nonnormalized_entry + QString( " (%1)").arg( total_song_count ) );

    last_item = entry_item;
    artist_count++;
  }

  setText( 0, text( 0 )+QString( " (%1)").arg( artist_count ) );
}




// tAlbumOverviewItem -------------------------------------------------------
tAlbumOverviewItem::tAlbumOverviewItem( tDatabase &database, QListView *view,  QListViewItem *after, const QString &label, const QString &criterion )
: tExpandableOverviewItem( view, after, label, criterion ), 
  Database( database )
{ 
}




tAlbumOverviewItem::tAlbumOverviewItem( tDatabase &database, QListViewItem *parent, QListViewItem *after, const QString &label, const QString &criterion )
: tExpandableOverviewItem( parent, after, label, criterion ),
  Database( database )
{ 
}




void tAlbumOverviewItem::addExpansion()
{
  tSongList songs( Database.SongCollection.begin(), Database.SongCollection.end() );
  ::sort( songs.begin(), songs.end(), less_song( FIELD_ALBUM ) );

  int album_count = 0;
  int song_count = 0;

  QListViewItem *last_item = NULL;
  QString last_entry = QString::null;
  FOREACH( first, songs, tSongList )
  {
    QString this_nonnormalized_entry = (*first)->album();
    QString this_entry = normalize( this_nonnormalized_entry );
    if ( this_nonnormalized_entry == "" )
      this_nonnormalized_entry = "<none>";

    if ( this_entry != last_entry )
    {
      if ( last_item )
      {
	last_item->setText( 0, last_item->text( 0 ) + QString(" (%1)").arg( song_count ) );
	song_count = 0;
      }

      last_item = new tOverviewItem( this, last_item, this_nonnormalized_entry, QString( "~album(complete:%1)" ).arg( quoteString( this_entry ) ) );
      album_count++;
    }
    last_entry = this_entry;
    song_count++;
  }

  if ( last_item )
    last_item->setText( 0, last_item->text( 0 ) + QString(" (%1)").arg( song_count ) );

  setText( 0, text( 0 )+QString( " (%1)").arg( album_count ) );
}





// tGenreOverviewItem -------------------------------------------------------
tGenreOverviewItem::tGenreOverviewItem( tDatabase &database, QListView *view,  QListViewItem *after, const QString &label, const QString &criterion )
: tExpandableOverviewItem( view, after, label, criterion ), 
  Database( database )
{ 
}




tGenreOverviewItem::tGenreOverviewItem( tDatabase &database, QListViewItem *parent, QListViewItem *after, const QString &label, const QString &criterion )
: tExpandableOverviewItem( parent, after, label, criterion ),
  Database( database )
{ 
}




void tGenreOverviewItem::addExpansion()
{
  tSongList songs( Database.SongCollection.begin(), Database.SongCollection.end() );

  ::sort( songs.begin(), songs.end(), less_song( FIELD_GENRE ) );
  QListViewItem *last_item = NULL;
  QString last_entry = QString::null;
  int song_count = 0;
  int genre_count = 0;

  FOREACH( first, songs, tSongList )
  {
    QString this_nonnormalized_entry = (*first)->genre();
    QString this_entry = normalize( this_nonnormalized_entry );
    if ( this_nonnormalized_entry == "" )
      this_nonnormalized_entry = "<none>";
    if ( this_entry != last_entry )
    {
      if ( last_item )
      {
	last_item->setText( 0, last_item->text( 0 ) + QString(" (%1)").arg( song_count ) );
	song_count = 0;
      }

      last_item = new tOverviewItem( this, last_item, this_nonnormalized_entry, QString( "~genre(complete:%1)" ).arg( quoteString( this_entry ) ) );
      genre_count++;
    }
    last_entry = this_entry;
    song_count++;
  }

  if ( last_item )
    last_item->setText( 0, last_item->text( 0 ) + QString(" (%1)").arg( song_count ) );

  setText( 0, text( 0 )+QString( " (%1)").arg( genre_count ) );
}




// tOverviewManager -----------------------------------------------------------
tOverviewManager::tOverviewManager( QListView &listview, tMainWindow *mw )
  : ListView( listview ), MainWindow( mw )
{
  connect( &ListView, SIGNAL( contextMenuRequested( QListViewItem *, const QPoint &, int ) ),
           this, SLOT( contextMenuRequested( QListViewItem *, const QPoint &, int ) ) );
  Menu = buildPopupMenu();
}




QPopupMenu *tOverviewManager::buildPopupMenu()
{
  QPopupMenu *menu = new QPopupMenu();

  QPopupMenu *play_menu = new QPopupMenu( &ListView );
  play_menu->insertItem( tr( "&Play now" ), this, SLOT( slotPlayNow() ) );
  play_menu->insertItem( tr( "Play &next" ), this, SLOT( slotPlayNext() ) );
  play_menu->insertItem( tr( "Play &evenutally" ), this, SLOT( slotPlayEventually() ) );
  menu->insertItem( tr( "&Play" ), play_menu );

  return menu;
}




void tOverviewManager::contextMenuRequested ( QListViewItem * item, const QPoint & pos, int col )
{
  Menu->popup( pos );
}




void tOverviewManager::slotPlayNow()
{
  tOverviewItem *my_item = dynamic_cast<tOverviewItem *>( ListView.selectedItem() );
  if ( my_item )
  {
    tSongSet set;
    set.setSongCollection( &MainWindow->database().SongCollection );
    set.setCriterion( my_item->criterion() );
    set.reevaluateCriterion();

    tSongList list;
    set.render( list );

    if ( MainWindow->preferences().Player )
      MainWindow->preferences().Player->playNow( list );
  }
}




void tOverviewManager::slotPlayNext()
{
  tOverviewItem *my_item = dynamic_cast<tOverviewItem *>( ListView.selectedItem() );
  if ( my_item )
  {
    tSongSet set;
    set.setSongCollection( &MainWindow->database().SongCollection );
    set.setCriterion( my_item->criterion() );
    set.reevaluateCriterion();

    tSongList list;
    set.render( list );

    if ( MainWindow->preferences().Player )
      MainWindow->preferences().Player->playNext( list );
  }
}




void tOverviewManager::slotPlayEventually()
{
  tOverviewItem *my_item = dynamic_cast<tOverviewItem *>( ListView.selectedItem() );
  if ( my_item )
  {
    tSongSet set;
    set.setSongCollection( &MainWindow->database().SongCollection );
    set.setCriterion( my_item->criterion() );
    set.reevaluateCriterion();

    tSongList list;
    set.render( list );

    if ( MainWindow->preferences().Player )
      MainWindow->preferences().Player->playEventually( list );
  }
}




