/*
madman - a music manager
Copyright (C) 2003  Andreas Kloeckner <ak@ixion.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/




#include <qlistbox.h>
#include <qpushbutton.h>
#include <qcombobox.h>
#include <qcheckbox.h>
#include <qtabwidget.h>
#include <qspinbox.h>
#include <qlistview.h>
#include <qlabel.h>
#include <qslider.h>

#include <stdexcept>

#include "utility/player.h"
#include "utility/plugin.h"
#include "ui/mainwin.h"
#include "ui/prefs_dialog.h"

#include "prefs_dialog.moc"




// tPreferencesDialog ---------------------------------------------------------



tPreferencesDialog::tPreferencesDialog( QWidget* parent, tDirectoryList &media_dir_list, 
    tDirectoryList &plugin_dir_list, QSettings &settings )
: tPreferencesDialogBase( parent, "prefs", true ),
  MediaDirectoryListManager( this, media_dir_list, lstMediaDirectories, btnAddMediaDirectory, btnRemoveMediaDirectory ),
  PluginDirectoryListManager( this, plugin_dir_list, lstPluginDirectories, btnAddPluginDirectory, btnRemovePluginDirectory ),
  Settings( settings )
{
  connect( &PluginDirectoryListManager, SIGNAL( changed() ),
      this, SLOT( updatePluginList() ) );
  connect( lstPlugins, SIGNAL( highlighted( int ) ),
      this, SLOT( updateSelectedPluginInfo( int ) ) );
  connect( btnConfigurePlugin, SIGNAL( clicked() ),
      this, SLOT( configurePlugin() ) );
  updatePluginList();
}




void tPreferencesDialog::updatePluginList()
{
  tPluginList plugins;
  enumeratePlugins( PluginDirectoryListManager.directoryList(),
      plugins );

  lstPlugins->clear();
  FOREACH_CONST( first, plugins, tPluginList )
    lstPlugins->insertItem( first->Name );
}




void tPreferencesDialog::updateSelectedPluginInfo( int index )
{
  tPluginList plugins;
  enumeratePlugins( PluginDirectoryListManager.directoryList(),
      plugins );

  if ( index < (int) plugins.size() )
  {
    SelectedPlugin = plugins[ index ].Filename;
    lblPluginName->setText( plugins[ index ].Name );
    lblPluginDescription->setText( plugins[ index ].Description );
  }
}




void tPreferencesDialog::configurePlugin()
{
  tPluginList plugins;
  enumeratePlugins( PluginDirectoryListManager.directoryList(),
      plugins );
  FOREACH( first, plugins, tPluginList )
    if ( first->Filename == SelectedPlugin )
    {
      first->configure( Settings );
      return;
    }
}




// public functions -----------------------------------------------------------
bool editPreferences( tMainWindow *mainwin, tPreferences &prefs, 
    tDirectoryList &media_dir_list, 
    QSettings &settings, int starting_tab )
{
  // create the dialog
  tDirectoryList work_media_dir_list,work_plugin_dir_list;
  work_media_dir_list = media_dir_list;
  work_plugin_dir_list = prefs.PluginDirectories;

  tPreferencesDialog prefsdialog( mainwin, 
      work_media_dir_list, 
      work_plugin_dir_list, settings );
  prefsdialog.tabMain->setCurrentPage( starting_tab );

  // populate the dialog
  vector<QString> players;
  listPlayers( players );
  int index = 0,current_index = 0;
  FOREACH_CONST( first, players, vector<QString> )
  {
    if ( prefs.Player && prefs.Player->name() == *first )
      current_index = index;
    prefsdialog.comboPlayer->insertItem( *first );
    index++;
  }
  prefsdialog.comboPlayer->setCurrentItem( current_index );

#define SET_BOOL( NAME ) \
  prefsdialog.check##NAME->setChecked(  prefs.NAME );

  SET_BOOL( ScanAtStartup );
  SET_BOOL( RememberGeometry );
  SET_BOOL( CollectHistory );
  SET_BOOL( UseFuzzySearch );

  prefsdialog.spinBackupCount->setValue( prefs.BackupCount );

  prefsdialog.lstVisibleColumns->setSorting( -1 );
  QCheckListItem *last = NULL;
  for ( int i = 0; i < FIELD_COUNT; i++ )
  {
    last = new QCheckListItem( prefsdialog.lstVisibleColumns, last,
	getFieldName( (tSongField) i ), QCheckListItem::CheckBox );
    last->setOn( prefs.ColumnVisibility[ i ] );
  }

  SET_BOOL( HttpDaemonEnabled );
  prefsdialog.spinHttpDaemonPort->setValue( prefs.HttpDaemonPort );
  SET_BOOL( HttpDownloadsEnabled );
  SET_BOOL( HttpScriptingEnabled );
  SET_BOOL( HttpWriteScriptingEnabled );
  SET_BOOL( HttpLocalPlayEnabled );
  SET_BOOL( HttpBrowsingEnabled );

  prefsdialog.sliderRatioUnrated->setValue( prefs.AutoDJPreferences.PlayRatioUnrated );
  prefsdialog.sliderRatio5Rated->setValue( prefs.AutoDJPreferences.PlayRatio5Rated );
  prefsdialog.sliderRatio4Rated->setValue( prefs.AutoDJPreferences.PlayRatio4Rated );
  prefsdialog.sliderRatio3Rated->setValue( prefs.AutoDJPreferences.PlayRatio3Rated );
  prefsdialog.sliderRatio2Rated->setValue( prefs.AutoDJPreferences.PlayRatio2Rated );
  prefsdialog.sliderRatio1Rated->setValue( prefs.AutoDJPreferences.PlayRatio1Rated );
  prefsdialog.sliderRatio0Rated->setValue( prefs.AutoDJPreferences.PlayRatio0Rated );
  prefsdialog.sliderRatioUnplayed->setValue( prefs.AutoDJPreferences.PlayRatioUnplayed );
  prefsdialog.sliderRatioOftenPlayed->setValue( prefs.AutoDJPreferences.PlayRatioOftenPlayed );
  prefsdialog.sliderRatioRarelyPlayed->setValue( prefs.AutoDJPreferences.PlayRatioRarelyPlayed );
  prefsdialog.sliderRatioRandom->setValue( prefs.AutoDJPreferences.PlayRatioRandom );
  prefsdialog.sliderRatioUnratedByGoodArtist->setValue( prefs.AutoDJPreferences.PlayRatioUnratedByGoodArtist );

  // show
  int returncode = prefsdialog.exec();

  // process results
  if ( returncode == QDialog::Accepted )
  {
    media_dir_list = work_media_dir_list;
    prefs.PluginDirectories = work_plugin_dir_list;

    prefs.setPlayer( createPlayer( prefsdialog.comboPlayer->currentText() ) );

#define GET_BOOL( NAME ) \
    prefs.NAME = prefsdialog.check##NAME->isChecked();

    GET_BOOL( ScanAtStartup );
    GET_BOOL( RememberGeometry );
    GET_BOOL( CollectHistory );
    GET_BOOL( UseFuzzySearch );

    prefs.BackupCount = prefsdialog.spinBackupCount->value();

    QListViewItem *item = prefsdialog.lstVisibleColumns->firstChild();
    for ( int i = 0; i < FIELD_COUNT; i++ )
    {
      prefs.ColumnVisibility[ i ] = dynamic_cast<QCheckListItem *>(item)->isOn();
      item = item->nextSibling();
    }

    GET_BOOL( HttpDaemonEnabled );
    prefs.HttpDaemonPort = prefsdialog.spinHttpDaemonPort->value();
    GET_BOOL( HttpDownloadsEnabled );
    GET_BOOL( HttpScriptingEnabled );
    GET_BOOL( HttpWriteScriptingEnabled );
    GET_BOOL( HttpLocalPlayEnabled );
    GET_BOOL( HttpBrowsingEnabled );

    prefs.AutoDJPreferences.PlayRatioUnrated = prefsdialog.sliderRatioUnrated->value();
    prefs.AutoDJPreferences.PlayRatio5Rated = prefsdialog.sliderRatio5Rated->value();
    prefs.AutoDJPreferences.PlayRatio4Rated = prefsdialog.sliderRatio4Rated->value();
    prefs.AutoDJPreferences.PlayRatio3Rated = prefsdialog.sliderRatio3Rated->value();
    prefs.AutoDJPreferences.PlayRatio2Rated = prefsdialog.sliderRatio2Rated->value();
    prefs.AutoDJPreferences.PlayRatio1Rated = prefsdialog.sliderRatio1Rated->value();
    prefs.AutoDJPreferences.PlayRatio0Rated = prefsdialog.sliderRatio0Rated->value();
    
    prefs.AutoDJPreferences.PlayRatioUnplayed = prefsdialog.sliderRatioUnplayed->value();
    prefs.AutoDJPreferences.PlayRatioOftenPlayed = prefsdialog.sliderRatioOftenPlayed->value();
    prefs.AutoDJPreferences.PlayRatioRarelyPlayed = prefsdialog.sliderRatioRarelyPlayed->value();
    prefs.AutoDJPreferences.PlayRatioRandom = prefsdialog.sliderRatioRandom->value();
    prefs.AutoDJPreferences.PlayRatioUnratedByGoodArtist = prefsdialog.sliderRatioUnratedByGoodArtist->value();
    return true;
  }
  else
    return false;
}
