/*
madman - a music manager
Copyright (C) 2003  Andreas Kloeckner <ak@ixion.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/




#include <qlistview.h>
#include <qheader.h>
#include <qregexp.h>
#include <qmessagebox.h>
#include <qaccel.h>
#include <qdragobject.h>
#include <qdatetime.h>
#include <qtimer.h>
#include <qmessagebox.h>
#include <qfiledialog.h>
#include <qstatusbar.h>

#include <unistd.h>
#include <algorithm>

#include "utility/player.h"
#include "utility/plugin.h"
#include "utility/progress.h"
#include "utility/prefs.h"
#include "database/auto_dj.h"
#include "ui/song_set_view.h"
#include "ui/mainwin.h"

#include "song_set_view.moc"




// tools ----------------------------------------------------------------------
namespace
{
  tSong *itemToSong( QListViewItem *item )
  {
    return dynamic_cast<tSongListViewItem *>( item ) ->song();
  }
}




tFilename importM3UIntoSongSet( tSongSet *song_set )
{
  if ( song_set == NULL )
    return QString::null;
  QString filename = QFileDialog::getOpenFileName( QString::null,
      qApp->translate( "importM3UIntoSongSet", "Playlist (*.m3u)" ),
      qApp->mainWidget() );
  if ( filename.isNull() )
    return QString::null;

  QFile dbfile( filename );
  if ( dbfile.open( IO_ReadOnly ) )
  {
    QTextStream textstr( &dbfile );
    
    while ( !textstr.atEnd() )
    {
      QString line = textstr.readLine();
      if ( line[ 0 ] == '#' )
	continue;

      try
      {
	song_set->add( line );
      }
      catch (... )
      {
      }
    }
  }
  return filename;
}




// tSongListViewItem ---------------------------------------------------
tSongListViewItem::tSongListViewItem( tSongSetViewManager &mgr, 
    tSong *song, float match, QListViewItem *after )
: QListViewItem( &mgr.ListView, after ), Song( song ), Manager( mgr ),
  CriterionMatch( match )
{
  setDragEnabled( true );

  setRenameEnabledInternal( FIELD_ARTIST, true );
  setRenameEnabledInternal( FIELD_PERFORMER, true );
  setRenameEnabledInternal( FIELD_TITLE, true );
  setRenameEnabledInternal( FIELD_ALBUM, true );
  setRenameEnabledInternal( FIELD_GENRE, true );
  setRenameEnabledInternal( FIELD_TRACKNUMBER, true );
  setRenameEnabledInternal( FIELD_YEAR, true );
}




void tSongListViewItem::setRenameEnabledInternal( int field, bool value )
{
  if ( Manager.FieldToColumn[ field ] >= 0 )
    setRenameEnabled( Manager.FieldToColumn[ field ], value );

}




tSong *tSongListViewItem::song()
{
  return Song;
}




QString tSongListViewItem::text ( int column ) const
{
  try
  {
    column = Manager.ColumnToField[ column ];
    switch ( column )
    {
      case FIELD_ARTIST:
	return Song->artist();
      case FIELD_PERFORMER:
	return Song->performer();
      case FIELD_TITLE:
	return Song->title();
      case FIELD_ALBUM:
	return Song->album();
      case FIELD_TRACKNUMBER:
	return Song->trackNumber();
      case FIELD_DURATION:
	{
	  int total_seconds = (int) Song->duration();
	  int seconds = total_seconds % 60;
	  int minutes = total_seconds / 60;

	  QString duration;
	  duration.sprintf( "%d:%02d", minutes, seconds );
	  return duration;
	}
      case FIELD_GENRE:
	return Song->genre();
      case FIELD_FULLPLAYCOUNT:
	return QString( "%1" ).arg( Song->fullPlayCount() );
      case FIELD_PARTIALPLAYCOUNT:
	return QString( "%1" ).arg( Song->partialPlayCount() );
      case FIELD_PLAYCOUNT:
	return QString( "%1" ).arg( Song->playCount() );

      case FIELD_LASTPLAYED:
	{
	  if ( Song->playCount() )
	  {
	    QDateTime datetime;
	    datetime.setTime_t( Song->lastPlayed() );
	    return datetime.toString( "MMM d h:mm" );
	  }
	  else
	    return "";
	}

      case FIELD_RATING:
	{
	  if ( Song->rating() == 0 )
	  {
	    return "-";
	  }
	  else if ( Song->rating() > 0 )
	  {
	    QString rating;
	    rating.fill( '*', Song->rating() );
	    return rating;
	  }
	  else
	    return "";
	}

      case FIELD_YEAR:
	return Song->year();
      case FIELD_FILE:
	return Song->filenameOnly();
      case FIELD_PATH:
	return Song->pathname();
      case FIELD_SIZE:
	return tr( "%1 MB" ).arg( double( Song->fileSize() ) / (1024 * 1024 ), 0, 'f', 2 );

      case FIELD_EXISTS_SINCE:
	{
	  QDateTime datetime;
	  datetime.setTime_t( Song->existsSince() );
	  return datetime.toString( "MMM d h:mm" );
	}
      case FIELD_EXTRA_CRITERION_MATCH:
	return QString::number( int( CriterionMatch * 100 ) ) + "%";
      default:
	return "";
    }
  }
  catch ( exception &ex )
  {
    return tr( "<ERROR>" );

  }
}




void tSongListViewItem::update()
{
  listView()->triggerUpdate();
}




void tSongListViewItem::noticeSongModified( tSong *song )
{
  if ( Song == song )
    update();
}




// tSongListView --------------------------------------------------------------
tSongListView::tSongListView( QWidget *parent, const char *name, WFlags f )
: tAcceleratorListView( parent, name, f )
{
}




tSong *tSongListView::highlightedSong()
{
  QListViewItem * item = currentItem();
  if ( !item )
    return NULL;
  return itemToSong( item );
}




void tSongListView::getSelection( tSongList &songlist )
{
  songlist.clear();
  QListViewItem *item = firstChild();
  while ( item )
  {
    if ( item->isSelected() )
      songlist.push_back( itemToSong( item ) );
    item = item->nextSibling();
  }
}




void tSongListView::getSelection( tIndexList &indexlist )
{
  indexlist.clear();
  int index = 0;

  QListViewItem *item = firstChild();
  while ( item )
  {
    if ( item->isSelected() )
      indexlist.push_back( index );
    item = item->nextSibling();
    index++;
  }

}




void tSongListView::setSongDropEnable( bool enabled )
{
  setAcceptDrops( true );
  viewport()->setAcceptDrops( true );
}




QDragObject *tSongListView::dragObject()
{
  QStrList uris;
  tSongList songlist;
  getSelection( songlist );

  FOREACH_CONST( first, songlist, tSongList )
    uris.append( QUriDrag::localFileToUri( (*first)->filename() ) );

  return new QUriDrag( uris, this );
}



void tSongListView::contentsDragEnterEvent( QDragEnterEvent *event )
{
  if ( QUriDrag::canDecode( event ) )
    event->accept();
}




void tSongListView::contentsDragMoveEvent( QDragMoveEvent *event )
{
  if ( QUriDrag::canDecode( event ) )
    event->accept();
}




void tSongListView::contentsDropEvent( QDropEvent *event )
{
  if ( QUriDrag::canDecode( event ) )
  {
    QStrList uri_list;
    QUriDrag::decode( event, uri_list );

    vector<tFilename> filenames;
    char *uri = uri_list.first();
    while ( uri )
    {
      QString filename = QUriDrag::uriToLocalFile( uri );
      if ( !filename.isNull() )
	filenames.push_back( filename );
      uri = uri_list.next();
    }

    tSongListViewItem *item = (tSongListViewItem *) itemAt( 
	contentsToViewport( event->pos() ) );
    tSong *onto_song = NULL;
    if ( item ) 
      onto_song = item->song();
    
    emit songsDropped( event->source() == this, onto_song, filenames );
  }
}




// tSongRetag -----------------------------------------------------------------
void tSongRetag::executeOn( tSong *song )
{
  try
  {
    switch ( Column )
    {
      case FIELD_ARTIST:
	song->setArtist( NewText );
	break;
      case FIELD_PERFORMER:
	song->setPerformer( NewText );
	break;
      case FIELD_TITLE:
	song->setTitle( NewText );
	break;
      case FIELD_ALBUM:
	song->setAlbum( NewText );
	break;
      case FIELD_GENRE:
	song->setGenre( NewText );
	break;
      case FIELD_YEAR:
	song->setYear( NewText );
	break;
      case FIELD_TRACKNUMBER:
	song->setTrackNumber( NewText );
	break;
      default:
	break;
    }
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( qApp->mainWidget(), qApp->translate( "ErrorMessages", "madman" ),
	qApp->translate( "ErrorMessages", "Error while writing tag:\n%1" ).arg( ex.what() ), 
	QMessageBox::Ok, QMessageBox::NoButton );
  }
}




// tDeferredSongRetag ---------------------------------------------------------
void tDeferredSongRetag::execute()
{
  Retag.executeOn( Song );

  delete this;
}




// tSongSetViewManager --------------------------------------------------------
tSongSetViewManager::tSongSetViewManager( tSongListView &lv, tMainWindow &mainwin, 
    const QString & pref_prefix  )
: ListView( lv ), SongSet( NULL ), MainWindow( mainwin ),
  LastRetagValid( false ), DoDisplaySort( true ), SortField( FIELD_ARTIST ), 
  Ascending( true ), PreferencesPrefix( pref_prefix )
{
  connect( &ListView, SIGNAL( itemRenamed( QListViewItem *, int ) ),
           this, SLOT( itemRenamed( QListViewItem *, int ) ) );
  connect( &ListView, SIGNAL( contextMenuRequested( QListViewItem *, const QPoint &, int ) ),
           this, SLOT( contextMenuRequested( QListViewItem *, const QPoint &, int ) ) );
  connect( &ListView, SIGNAL( doubleClicked( QListViewItem * ) ),
           this, SLOT( slotPlayNow( QListViewItem * ) ) );
  connect( &ListView, SIGNAL( selectionChanged() ),
           this, SLOT( selectionChanged() ) );
  connect( ListView.header(), SIGNAL( clicked( int ) ),
           this, SLOT( headerClicked( int ) ) );
}




void tSongSetViewManager::setup()
{
  auto_ptr<QPopupMenu> new_menu ( buildPopupMenu() );
  Menu = new_menu;
}




void tSongSetViewManager::setupListView( const vector<bool> &visibility )
{
  FieldToColumn.resize( FIELD_COUNT );
  ColumnToField.resize( FIELD_COUNT );

  ListView.setSorting( -1 );

  ListView.clear();
  while ( ListView.columns() )
    ListView.removeColumn( 0 );

  int column = 0;

  for ( int field = 0; field < FIELD_COUNT; field++ )
  {
    if ( visibility[ field ] )
    {
      ListView.addColumn( getFieldName( (tSongField) field ) );
      ListView.setColumnWidthMode( column, QListView::Manual );

      FieldToColumn[ field ] = column;
      ColumnToField[ column ] = field;
      column++;
    }
    else
    {
      FieldToColumn[ field ] = -1;
    }
  }

  ListView.setResizeMode( QListView::NoColumn );
  ListView.setSelectionMode( QListView::Extended );
  ListView.setAllColumnsShowFocus( true );
}




void tSongSetViewManager::setSongSet( tSongSet *ss )
{
  if ( SongSet )
  {
    disconnect( SongSet, NULL, this, NULL );
  }
  SongSet = ss;
  if ( SongSet )
  {
    connect( SongSet, SIGNAL( notifyChange( bool, tProgress *) ), this, SLOT( noticeSongSetModified( bool, tProgress * ) ) );
    connect( SongSet, SIGNAL( notifySongModified( tSong *) ), this, SLOT( noticeSongModified ( tSong * ) ) );
  }

  redisplay( false );
}




QPopupMenu *tSongSetViewManager::buildPopupMenu()
{
  QPopupMenu * menu = new QPopupMenu( &ListView );

  // Play menu ----------------------------------------------------------------
  QPopupMenu *play_menu = new QPopupMenu( &ListView );
  play_menu->insertItem( tr( "&Play now (F9)" ), this, SLOT( slotPlayNow() ) );
  play_menu->insertItem( tr( "Play &next (F10)" ), this, SLOT( slotPlayNext() ) );
  play_menu->insertItem( tr( "Play &evenutally (F11)" ), this, SLOT( slotPlayEventually() ) );
  play_menu->insertSeparator();
  play_menu->insertItem( tr( "Auto DJ 20 songs" ), this, SLOT( slotAutoDJ() ) );

  menu->insertItem( tr( "&Play" ), play_menu );

  ListView.addShortCut( new tKeyboardShortCut( Key_F9, this, SLOT( slotPlayNow() ) ) );
  ListView.addShortCut( new tKeyboardShortCut( Key_F10, this, SLOT( slotPlayNext() ) ) );
  ListView.addShortCut( new tKeyboardShortCut( Key_F11, this, SLOT( slotPlayEventually() ) ) );

  // Tag menu -----------------------------------------------------------------
  QPopupMenu *tag_menu = new QPopupMenu( &ListView );
  tag_menu->insertItem( tr( "From Filename: &Artist - Title" ), this, SLOT( tagArtistTitle() ) );
  tag_menu->insertItem( tr( "From Filename: &Title - Artist" ), this, SLOT( tagTitleArtist() ) );
  tag_menu->insertSeparator();
  tag_menu->insertItem( tr( "Repeat last retag on selection" ), this, SLOT( redoRetag() ) );
  tag_menu->insertItem( tr( "R&eread tag" ), this, SLOT( rereadTag() ) );
  tag_menu->insertItem( tr( "Re&write tag" ), this, SLOT( rewriteTag() ) );
  tag_menu->insertItem( tr( "&Strip tag" ), this, SLOT( stripTag() ) );

  menu->insertSeparator();
  menu->insertItem( tr( "&Tag" ), tag_menu );

  // Rate menu ----------------------------------------------------------------
  QPopupMenu *rate_menu = new QPopupMenu( &ListView );
  rate_menu->insertItem( tr( "&No rating" ), this, SLOT( rateNone() ) );
  rate_menu->insertSeparator();
  rate_menu->insertItem( tr( "&0: -" ), this, SLOT( rate0() ) );
  rate_menu->insertItem( tr( "&1: *" ), this, SLOT( rate1() ) );
  rate_menu->insertItem( tr( "&2: **" ), this, SLOT( rate2() ) );
  rate_menu->insertItem( tr( "&3: ***" ), this, SLOT( rate3() ) );
  rate_menu->insertItem( tr( "&4: ****" ), this, SLOT( rate4() ) );
  rate_menu->insertItem( tr( "&5: *****" ), this, SLOT( rate5() ) );

  menu->insertSeparator();
  menu->insertItem( tr( "&Rate" ), rate_menu );

  ListView.addShortCut( new tKeyboardShortCut( Key_0, this, SLOT( rate0() ) ) );
  ListView.addShortCut( new tKeyboardShortCut( Key_1, this, SLOT( rate1() ) ) );
  ListView.addShortCut( new tKeyboardShortCut( Key_2, this, SLOT( rate2() ) ) );
  ListView.addShortCut( new tKeyboardShortCut( Key_3, this, SLOT( rate3() ) ) );
  ListView.addShortCut( new tKeyboardShortCut( Key_4, this, SLOT( rate4() ) ) );
  ListView.addShortCut( new tKeyboardShortCut( Key_5, this, SLOT( rate5() ) ) );

  // Plugins menu -------------------------------------------------------------
  QPopupMenu *plugins_menu = new QPopupMenu( &ListView );
  MenuIdToPluginFilenameMap.clear();
  tPluginList plugins;
  enumeratePlugins( MainWindow.preferences().PluginDirectories, plugins );
  FOREACH( first, plugins, tPluginList )
  {
    int id = plugins_menu->insertItem( first->MenuString );
    MenuIdToPluginFilenameMap[ id ] = first->Filename;
  }
  connect( plugins_menu, SIGNAL( activated(int) ),
      this, SLOT( pluginActivated( int ) ) );

  menu->insertSeparator();
  menu->insertItem( tr( "Run &Plugin" ), plugins_menu );

  // Others -------------------------------------------------------------------
  menu->insertSeparator();
  menu->insertItem( tr( "&Delete from disk (Ctrl+Del)" ), this, SLOT( deleteFromDisk() ) );
  menu->insertItem( tr( "R&eset play statistics" ), this, SLOT( resetStatistics() ) );
  menu->insertItem( tr( "&Highlight current song (F12)" ), this, SLOT( highlightCurrentSong() ) );

  ListView.addShortCut( new tKeyboardShortCut( Key_F12, this, SLOT( highlightCurrentSong() ) ) );
  ListView.addShortCut( new tKeyboardShortCut( CTRL + Key_Delete, this, SLOT( deleteFromDisk() ) ) );

  return menu;
}




void tSongSetViewManager::saveListViewAppearance( QSettings &settings )
{
  for ( int field = 0; field < FIELD_COUNT; field++ )
  {
    if ( FieldToColumn[ field ] == -1 )
      continue;

    int column = FieldToColumn[ field ];

    settings.writeEntry( QString( "%1/listview.columns.%2.width" ).
	arg( PreferencesPrefix ).arg( field ),
	ListView.columnWidth( column ) );
    settings.writeEntry( QString( "%1/listview.columns.%2.display_index" ).
	arg( PreferencesPrefix ).arg( field ),
	ListView.header()->mapToIndex( column ) );
  }
}




void tSongSetViewManager::loadListViewAppearance( QSettings &settings )
{
  // FIXME poor man's sorting, dumb algorithm
  for ( int current_column = 0; current_column < FIELD_COUNT; current_column++ )
  {
    for ( int field = 0; field < FIELD_COUNT; field++ )
    {
      int column = FieldToColumn[ field ];
      if ( column == -1 )
	continue;

      bool ok;
      int display_index = settings.readNumEntry( 
	  QString( "%1/listview.columns.%2.display_index" ).
	  arg( PreferencesPrefix ).arg( field ), 0, &ok ); 
      if ( ok )
      {
	if ( current_column != display_index )
	  continue;
	ListView.header()->moveSection( column, display_index );

	int width = settings.readNumEntry( 
	    QString( "%1/listview.columns.%2.width" ).
	    arg( PreferencesPrefix ).arg( field ), 0, &ok ); 
	if ( ok )
	  ListView.setColumnWidth( column, width );
      }
    }
  }
}




void tSongSetViewManager::noticeSongModified( tSong *song )
{
  QListViewItem *item = ListView.firstChild();

  while ( item )
  {
    dynamic_cast<tSongListViewItem *>( item )->noticeSongModified( song );
    item = item->nextSibling();
  }
}




void tSongSetViewManager::noticeSongSetModified( bool substantial_change, tProgress *progress )
{
  redisplay( !substantial_change, progress );
}




void tSongSetViewManager::redisplay( bool preserve_scroll_pos, tProgress *progress )
{
  if ( progress )
    progress->setWhat( tr( "Displaying..." ) );

  ListView.setUpdatesEnabled( false );
  ListView.viewport()->setUpdatesEnabled( false );
  ListView.header()->setUpdatesEnabled( false );

  int contents_x = ListView.contentsX();
  int contents_y = ListView.contentsY();

  ListView.clear();

  if ( SongSet )
  {
    tRendering list;
    SongSet->render( list );

    if ( DoDisplaySort )
    {
      sort( list.begin(), list.end(), less_rendering( SortField, Ascending ) );
      
      if ( FieldToColumn[ SortField ] >= 0 )
	ListView.header()->setSortIndicator( FieldToColumn[ SortField ], Ascending );
      else
	ListView.header()->setSortIndicator( -1 );
    }

    QListViewItem *last_item = NULL;

    FOREACH_CONST( first, list, tRendering )
    {
      last_item = new tSongListViewItem( *this, 
	  first->Song, first->CriterionMatch, last_item );

      if ( progress )
      {
	progress->processEvents();
	if ( progress->wasCancelled() )
	  return;
      }
    }
  }

  if ( preserve_scroll_pos )
    ListView.setContentsPos( contents_x, contents_y );

  ListView.setUpdatesEnabled( true );
  ListView.viewport()->setUpdatesEnabled( true );
  ListView.header()->setUpdatesEnabled( true );

  ListView.triggerUpdate();
  ListView.viewport()->update();
  ListView.header()->update();
}





void tSongSetViewManager::itemRenamed( QListViewItem *item, int col )
{
  tSongListViewItem *cast_item = dynamic_cast<tSongListViewItem *>( item );
  QString new_text = cast_item->QListViewItem::text( col );
  tSong *song = cast_item->song();

  if ( new_text == cast_item->text( col ) )
    return;

  tDeferredSongRetag *retag = new tDeferredSongRetag;
  retag->Song = song;
  retag->Retag.NewText = new_text;
  retag->Retag.Column = (tSongField) ColumnToField[ col ];
  QTimer::singleShot( 0, retag, SLOT( execute() ) );

  LastRetagValid = true;
  LastRetag = retag->Retag;
}




void tSongSetViewManager::contextMenuRequested( QListViewItem *item, const QPoint &pos, int col )
{
  Menu->popup( pos );
}




void tSongSetViewManager::slotPlayNow()
{
  tSongList list;
  ListView.getSelection( list );
  if ( MainWindow.preferences().Player )
    MainWindow.preferences().Player->playNow( list );
}




void tSongSetViewManager::slotPlayNow( QListViewItem *item )
{
  ListView.clearSelection();
  ListView.setSelected( item, true );

  tSongList list;
  ListView.getSelection( list );

  if ( MainWindow.preferences().Player )
    MainWindow.preferences().Player->playNow( list );
}




void tSongSetViewManager::slotPlayNext()
{
  tSongList list;
  ListView.getSelection( list );

  if ( MainWindow.preferences().Player )
    MainWindow.preferences().Player->playNext( list );
}




void tSongSetViewManager::slotPlayEventually()
{
  tSongList list;
  ListView.getSelection( list );

  if ( MainWindow.preferences().Player )
    MainWindow.preferences().Player->playEventually( list );
}




void tSongSetViewManager::slotAutoDJ()
{
  try
  {
    if ( SongSet == NULL )
      return;

    tSongList songs;
    doAutoDJ( MainWindow.database().SongCollection, 
	SongSet, songs, MainWindow.preferences().AutoDJPreferences, 20 );

    if ( MainWindow.preferences().Player )
      MainWindow.preferences().Player->playEventually( songs );
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( &ListView, tr( "madman" ),
	tr( "Auto DJ encountered an error:\n%1" ).arg( ex.what() ), QMessageBox::Ok, QMessageBox::NoButton );
  }
}




void tSongSetViewManager::rate( int rating )
{
  tSongList list;
  ListView.getSelection( list );

  FOREACH( first, list, tSongList )
    (*first)->rate( rating );
}




void tSongSetViewManager::resetStatistics()
{
  tSongList list;
  ListView.getSelection( list );

  FOREACH( first, list, tSongList )
    (*first)->resetStatistics();
}




void tSongSetViewManager::deleteFromDisk()
{
  try
  {
    tSongList list;
    ListView.getSelection( list );

    if ( QMessageBox::warning( &ListView, tr( "File Deletion" ),
	  tr( "OK to physically delete %1 files?" ).arg( list.size() ),
	  QMessageBox::Yes, QMessageBox::No ) == QMessageBox::Yes )
    {
      MainWindow.database().SongCollection.beginBulkChange();
      FOREACH_CONST( first, list, tSongList )
      {
	if ( MainWindow.database().SongCollection.isValid( *first ) )
	{
	  MainWindow.database().SongCollection.remove( *first );
	  unlink( ( *first ) ->filename().latin1() );
	}
      }
      MainWindow.database().SongCollection.endBulkChange();
    }
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( &ListView, tr( "madman" ),
	tr( "Error while deleting file:\n%1" ).arg( ex.what() ), QMessageBox::Ok, QMessageBox::NoButton );
  }
}




void tSongSetViewManager::highlightCurrentSong()
{
  if ( MainWindow.preferences().Player == NULL )
    return;

  tFilename current_song = MainWindow.preferences().Player->currentFilename();
  tSong *song = MainWindow.database().SongCollection.getByFilename( current_song );
  if ( song )
    highlight( song );
}




void tSongSetViewManager::tagArtistTitle()
{
  try
  {
    tSongList list;
    ListView.getSelection( list );

    QRegExp dash_pattern( "^([^-]+)\\s*\\-\\s*(.+)\\.(mp3|ogg)$" );
    dash_pattern.setCaseSensitive( false );
    dash_pattern.setMinimal( true );

    FOREACH_CONST( first, list, tSongList )
    {
      if ( dash_pattern.exactMatch( (*first)->filenameOnly() ) )
      {
	(*first)->setArtist( dash_pattern.cap( 1 ).stripWhiteSpace() );
	(*first)->setTitle( dash_pattern.cap( 2 ).stripWhiteSpace() );
      }
    }
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( &ListView, tr( "madman" ),
	tr( "Error while writing tag:\n%1" ).arg( ex.what() ), QMessageBox::Ok, QMessageBox::NoButton );
  }
}




void tSongSetViewManager::tagTitleArtist()
{
  try
  {
    tSongList list;
    ListView.getSelection( list );

    QRegExp dash_pattern( "^([^-]+)\\s*\\-\\s*(.+)\\.(mp3|ogg)$" );
    dash_pattern.setCaseSensitive( false );
    dash_pattern.setMinimal( true );
    FOREACH_CONST( first, list, tSongList )
    {
      if ( dash_pattern.exactMatch( (*first)->filenameOnly() ) )
      {
	(*first)->setTitle( dash_pattern.cap( 1 ).stripWhiteSpace() );
	(*first)->setArtist( dash_pattern.cap( 2 ).stripWhiteSpace() );
      }
    }
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( &ListView, tr( "madman" ),
	tr( "Error while writing tag:\n%1" ).arg( ex.what() ), QMessageBox::Ok, QMessageBox::NoButton );
  }
}




void tSongSetViewManager::redoRetag()
{
  if ( LastRetagValid )
  {
    tSongList list;
    ListView.getSelection( list );

    if ( QMessageBox::warning( &ListView, tr( "madman" ),
	  tr( "OK to set the '%1' column of %2 songs to '%3'?")
	  .arg( getFieldName( LastRetag.Column ) )
	  .arg( list.size() )
	  .arg( LastRetag.NewText ) ,
	  QMessageBox::Yes, QMessageBox::No ) == QMessageBox::Yes )
    {
      FOREACH_CONST( first, list, tSongList )
	LastRetag.executeOn( *first );
    }
  }
  else
    QMessageBox::information( &ListView, tr( "madman" ),
	tr( "There was no previous retag. Sorry." ), QMessageBox::Ok );
}




void tSongSetViewManager::rereadTag()
{
  tSongList list;
  ListView.getSelection( list );

  try
  {
    FOREACH_CONST( first, list, tSongList )
    {
      (*first)->invalidateCache();
      (*first)->ensureInfoIsThere();
    }
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( &ListView, tr( "madman" ),
	tr( "Error while rereading tag:\n%1" ).arg( ex.what() ), QMessageBox::Ok, QMessageBox::NoButton );
  }
}




void tSongSetViewManager::rewriteTag()
{
  tSongList list;
  ListView.getSelection( list );

  try
  {
    FOREACH_CONST( first, list, tSongList )
      (*first)->rewriteTag();
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( &ListView, tr( "madman" ),
	tr( "Error while rewriting tag:\n%1" ).arg( ex.what() ), QMessageBox::Ok, QMessageBox::NoButton );
  }
}




void tSongSetViewManager::stripTag()
{
  tSongList list;
  ListView.getSelection( list );

  try
  {
    FOREACH_CONST( first, list, tSongList )
      (*first)->stripTag();
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( &ListView, tr( "madman" ),
	tr( "Error while stripping tag:\n%1" ).arg( ex.what() ), QMessageBox::Ok, QMessageBox::NoButton );
  }
}




void tSongSetViewManager::removeSelected()
{
  if ( SongSet == NULL )
    return;

  tIndexList list;
  ListView.getSelection( list );
  sort( list.begin(), list.end() );
  reverse( list.begin(), list.end() );

  FOREACH( first, list, tIndexList )
    SongSet->remove( *first );
}




void tSongSetViewManager::searchForward( tCriterion *crit, bool from_beginning )
{
  tCriterion *criterion = crit;

  QListViewItem *item = ListView.firstChild();
  if ( !from_beginning && ListView.currentItem( ) )
    item = ListView.currentItem();

  if ( !item )
    return ;

  item = item->nextSibling();
  while ( item )
  {
    if ( ( *criterion ) ( itemToSong( item ) ) )
    {
      ListView.setCurrentItem( item );
      ListView.ensureItemVisible( item );
      ListView.clearSelection();
      ListView.setSelected( item, true );
      return;
    }
    item = item->nextSibling();
  }
}




void tSongSetViewManager::highlight( tSong *song )
{
  QListViewItem *item = ListView.firstChild();

  if ( !item )
    return ;

  while ( item )
  {
    if ( itemToSong( item ) == song )
    {
      ListView.setCurrentItem( item );
      ListView.ensureItemVisible( item );
      ListView.clearSelection();
      ListView.setSelected( item, true );
      return;
    }
    item = item->nextSibling();
  }
}




void tSongSetViewManager::setDisplaySorting( tSongField field, bool ascending )
{
  SortField = field;
  Ascending = ascending;
}




void tSongSetViewManager::pluginActivated( int id )
{
  try
  {
    tPluginList plugins;
    enumeratePlugins( MainWindow.preferences().PluginDirectories, plugins );
    hash_map<int,tFilename>::iterator it = MenuIdToPluginFilenameMap.find( id );
    if ( it != MenuIdToPluginFilenameMap.end() )
    {
      tFilename name = it->second;

      tSongList selection;
      ListView.getSelection( selection );

      FOREACH( first, plugins, tPluginList )
	if ( first->Filename == name )
	{
	  bool confirmed = !first->Confirm || QMessageBox::warning( &MainWindow, tr( "madman" ),
		tr( "Are you sure you want to run plugin '%1' on %2 songs?" )
		.arg( first->Name ).arg( selection.size() ),
		QMessageBox::Yes, QMessageBox::No ) == QMessageBox::Yes;

	  if ( confirmed )
	  {
	    first->run( MainWindow.settings(), selection );
	    if ( first->RescanAfter )
	      MainWindow.rescan();
	  }
	  return;
	}
    }
    else
      runtimeError( tr( "This plugin has mysteriously disappeared" ) );
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( &MainWindow, tr( "madman" ), 
	ex.what(), QMessageBox::Ok, 0 );
  }
}




void tSongSetViewManager::selectionChanged()
{
  tSongList list;
  ListView.getSelection( list );

  double duration = 0;
  double size = 0;
  FOREACH_CONST( first, list, tSongList )
  {
    duration += (*first)->duration();
    size += (*first)->fileSize() / (1024. * 1024 );
  }

  int days = int( duration / 3600. / 24. );
  duration -= days * 3600 * 24;
  int hours = int( duration / 3600 );
  duration -= hours * 3600;
  int minutes = int( duration / 60 );
  duration -= minutes * 60;
  int seconds = int( duration );

  QString duration_string;
  duration_string.sprintf( "%d:%02d", minutes, seconds );
  if ( days )
    duration_string.sprintf( "%d days %d:%02d:%02d", days, hours, minutes, seconds );
  else if ( hours )
    duration_string.sprintf( "%d:%02d:%02d", hours, minutes, seconds );
  else
    duration_string.sprintf( "%2d:%02d", minutes, seconds );

  QString message;
  if ( list.size() == 0)
  {
    message = tr( "0 songs selected." );

  }
  else if ( list.size() == 1)
  {
   message = tr( "%1 - [ Duration %2, Size %3 ]" )
      .arg( list[0]->filename() )
      .arg( duration_string )
      .arg( tr( "%1 MB" ).arg( size, 0, 'f', 2 ) );
  }
  else
  {
    message = tr( "%1 songs selected - Duration %2 - Size %3" )
      .arg( list.size() )
      .arg( duration_string )
      .arg( tr( "%1 MB" ).arg( size, 0, 'f', 2 ) );
  }

  MainWindow.statusBar()->message( message, 10000 );
}




// tEditableSongSetViewManager ------------------------------------------------
tEditableSongSetViewManager::tEditableSongSetViewManager( tSongListView &lv, tMainWindow &mainwin,
    const QString & pref_prefix )
: tSongSetViewManager( lv, mainwin, pref_prefix )
{
  ListView.setSongDropEnable( true );
  DoDisplaySort = false;

  connect( &ListView, SIGNAL( songsDropped(bool,tSong *,vector<tFilename> &) ),
	this, SLOT( songsDropped(bool,tSong *,vector<tFilename> &) ) );
}




QPopupMenu *tEditableSongSetViewManager::buildPopupMenu()
{
  QPopupMenu *menu = super::buildPopupMenu();

  menu->insertItem( tr( "Remo&ve from playlist (Del)" ), this, SLOT( removeSelected() ) );
  QPopupMenu *reorder_menu = new QPopupMenu( &ListView );
  reorder_menu->insertItem( tr( "Sort by &Artist" ), this, SLOT( sortByArtist() ) );
  reorder_menu->insertItem( tr( "Sort by &Title" ), this, SLOT( sortByTitle() ) );
  reorder_menu->insertItem( tr( "Sort by A&lbum" ), this, SLOT( sortByAlbum() ) );
  reorder_menu->insertItem( tr( "Sort by &Track number" ), this, SLOT( sortByTrackNumber() ) );
  reorder_menu->insertItem( tr( "Sort by &Year" ), this, SLOT( sortByYear() ) );
  reorder_menu->insertItem( tr( "Sort by &Genre" ), this, SLOT( sortByGenre() ) );
  reorder_menu->insertSeparator();
  reorder_menu->insertItem( tr( "Jumble" ), this, SLOT( jumble() ) );
  reorder_menu->insertItem( tr( "Reverse" ), this, SLOT( reverse() ) );
  menu->insertSeparator();
  menu->insertItem( tr( "&Reorder" ), reorder_menu );
  menu->insertSeparator();
  menu->insertItem( tr( "Export to player's current playlist" ), this, SLOT( uploadSongSet() ) );
  menu->insertItem( tr( "Import player's current playlist" ), this, SLOT( downloadSongSet() ) );
  menu->insertItem( tr( "Import M3U" ), this, SLOT( importM3U() ) );

  ListView.addShortCut( new tKeyboardShortCut( Key_Delete, this, SLOT( removeSelected() ) ) );

  return menu;
}




void tEditableSongSetViewManager::uploadSongSet()
{
  if ( SongSet )
  {
    tSongList rendering;
    SongSet->render( rendering );
    if ( MainWindow.preferences().Player )
      MainWindow.preferences().Player->setPlayList( rendering );
  }
}




void tEditableSongSetViewManager::downloadSongSet()
{
  if ( MainWindow.preferences().Player == NULL )
    return;
  if ( SongSet == NULL )
    return;

  vector<tFilename> playlist;
  MainWindow.preferences().Player->getPlayList( playlist );

  FOREACH( first, playlist, vector<tFilename> )
  {
    try
    {
      SongSet->add( *first );
    }
    catch (... )
    {
    }
  }
}




void tEditableSongSetViewManager::importM3U()
{
  importM3UIntoSongSet( SongSet );
}




void tEditableSongSetViewManager::sortBy( tSongField field )
{
  if ( SongSet )
    SongSet->sortBy( field );
}




void tEditableSongSetViewManager::jumble()
{
  if ( SongSet )
    SongSet->jumble();
}




void tEditableSongSetViewManager::reverse()
{
  if ( SongSet )
    SongSet->reverse();
}




void tEditableSongSetViewManager::songsDropped( bool from_same_widget, tSong *onto, vector<tFilename> &songs )
{
  if ( SongSet == NULL )
    return;

  int preferred_index = -1;

  SongSet->beginBulkChange();

  if ( from_same_widget )
  {
    removeSelected();
  }

  if ( onto )
  {
    tSongList rendering;
    SongSet->render( rendering );

    int index = 0;
    FOREACH_CONST( first, rendering, tSongList )
    {
      if ( *first == onto )
      {
	preferred_index = index;
	break;
      }
      index++;
    }
  }

  FOREACH_CONST( first, songs, vector<tFilename> )
  {
    try
    {
      SongSet->add( *first, preferred_index );
      if ( preferred_index != -1 )
	++preferred_index;
    }
    catch ( exception &ex )
    {
      qWarning( "%s", ex.what() );
    }
  }

  SongSet->endBulkChange();
}




void tEditableSongSetViewManager::headerClicked( int column )
{
  column = ColumnToField[ column ];
  sortBy( (tSongField) column );
}




// tAllSongsViewManager -------------------------------------------------------
tAllSongsViewManager::tAllSongsViewManager( tSongListView &lv, tMainWindow &mainwin, 
    const QString & pref_prefix  )
: tSongSetViewManager( lv, mainwin, pref_prefix )
{
  ListView.setShowSortIndicator( true );
}




QPopupMenu *tAllSongsViewManager::buildPopupMenu()
{
  QPopupMenu *menu = super::buildPopupMenu();

  menu->insertSeparator();
  menu->insertItem( tr( "&Add to current playlist (Ins)" ), this, SLOT( slotAdd() ) );

  ListView.addShortCut( new tKeyboardShortCut( Key_Insert, this, SLOT( slotAdd() ) ) );

  return menu;
}




void tAllSongsViewManager::slotAdd()
{
  tSongList list;
  ListView.getSelection( list );
  emit add( list );
}



void tAllSongsViewManager::headerClicked( int column )
{
  column = ColumnToField[ column ];

  if ( column == SortField )
  {
    Ascending = !Ascending;
  }
  else 
  {
    SortField = (tSongField) column;
    Ascending = true;
  }

  redisplay( true );
}






