/*
madman - a music manager
Copyright (C) 2003  Andreas Kloeckner <ak@ixion.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/




#include <qsettings.h>

#include <stdexcept>

#include "utility/player.h"
#include "utility/plugin.h"
#include "utility/prefs.h"
#include "database/song.h"




// tPreferences ---------------------------------------------------------------
void tPreferences::save( QSettings &settings )
{
  if ( Player )
  {
    settings.writeEntry( "/madman/player", Player->name() );
  }

#define WRITE_BOOL( SETTING, VARIABLE ) \
  settings.writeEntry( "/madman/" SETTING, VARIABLE ? 1 : 0 );

  QStringList plugin_dirs = QStringvector2QStringList( PluginDirectories );
  settings.writeEntry( "/madman/plugin_directories", plugin_dirs );
  WRITE_BOOL( "remember_geometry", RememberGeometry );
  WRITE_BOOL( "scan_at_startup", ScanAtStartup );
  WRITE_BOOL( "collect_history", CollectHistory );
  WRITE_BOOL( "use_fuzzy_search", UseFuzzySearch );

  settings.writeEntry( "/madman/number_of_backups_kept", BackupCount );

  for ( int i = 0; i < FIELD_COUNT; i++ )
    settings.writeEntry( 
	QString( "/madman/column.%1.visible" ).arg( i ), 
	ColumnVisibility[i] ? 1 : 0 );
  
  WRITE_BOOL( "http_daemon_enabled", HttpDaemonEnabled );
  settings.writeEntry( "/madman/http_server_port", HttpDaemonPort );
  WRITE_BOOL( "http_downloads_enabled", HttpDownloadsEnabled );
  WRITE_BOOL( "http_scripting_enabled", HttpScriptingEnabled );
  WRITE_BOOL( "http_write_scripting_enabled", HttpWriteScriptingEnabled );
  WRITE_BOOL( "http_local_play_enabled", HttpLocalPlayEnabled );
  WRITE_BOOL( "http_browsing_enabled", HttpBrowsingEnabled );

  settings.writeEntry( "/madman/auto_dj/play_ratio_unrated", AutoDJPreferences.PlayRatioUnrated );
  settings.writeEntry( "/madman/auto_dj/play_ratio_5_rated", AutoDJPreferences.PlayRatio5Rated );
  settings.writeEntry( "/madman/auto_dj/play_ratio_4_rated", AutoDJPreferences.PlayRatio4Rated );
  settings.writeEntry( "/madman/auto_dj/play_ratio_3_rated", AutoDJPreferences.PlayRatio3Rated );
  settings.writeEntry( "/madman/auto_dj/play_ratio_2_rated", AutoDJPreferences.PlayRatio2Rated );
  settings.writeEntry( "/madman/auto_dj/play_ratio_1_rated", AutoDJPreferences.PlayRatio1Rated );
  settings.writeEntry( "/madman/auto_dj/play_ratio_0_rated", AutoDJPreferences.PlayRatio0Rated );
  settings.writeEntry( "/madman/auto_dj/play_ratio_unplayed", AutoDJPreferences.PlayRatioUnplayed );
  settings.writeEntry( "/madman/auto_dj/play_ratio_often_played", AutoDJPreferences.PlayRatioOftenPlayed );
  settings.writeEntry( "/madman/auto_dj/play_ratio_rarely_played", AutoDJPreferences.PlayRatioRarelyPlayed );
  settings.writeEntry( "/madman/auto_dj/play_ratio_random", AutoDJPreferences.PlayRatioRandom );
  settings.writeEntry( "/madman/auto_dj/play_ratio_unrated_by_good_artist", AutoDJPreferences.PlayRatioUnratedByGoodArtist  );
}




void tPreferences::load( QSettings &settings )
{
  QString player_name = settings.readEntry( "/madman/player" );
  
  tPlayer * newplayer = createPlayer( player_name );
  setPlayer( newplayer );
  if ( !Player )
  {
    vector<QString> players;
    listPlayers( players );
    if ( players.size() )
      setPlayer( createPlayer( players[0] ) );
    if ( !Player )
      throw runtime_error( "can't instantiate any player, giving up." );
  }

  bool ok;
  PluginDirectories = QStringList2QStringvector( settings.readListEntry( "/madman/plugin_directories", &ok ) );
  if ( !ok )
  {
    PluginDirectories.clear();
    PluginDirectories.push_back( MADMAN_LIBDIR "/madman/plugins" );
  }
#define READ_BOOL( SETTING, VARIABLE, DEFAULT ) \
  VARIABLE = settings.readNumEntry( "/madman/" SETTING, DEFAULT );

  READ_BOOL( "remember_geometry", RememberGeometry, 1 );
  READ_BOOL( "scan_at_startup", ScanAtStartup, 0 );
  READ_BOOL( "collect_history", CollectHistory, 1 );
  READ_BOOL( "use_fuzzy_search", UseFuzzySearch, 0 );

  BackupCount = settings.readNumEntry( "/madman/number_of_backups_kept", 4 );

  ColumnVisibility.resize( FIELD_COUNT );
  for ( int i = 0; i < FIELD_COUNT; i++ )
  {
    int default_vis = 1;
    if ( i == FIELD_FULLPLAYCOUNT ) default_vis = 0;
    if ( i == FIELD_PARTIALPLAYCOUNT ) default_vis = 0;
    if ( i == FIELD_PERFORMER ) default_vis = 0;
    if ( i == FIELD_EXTRA_CRITERION_MATCH ) default_vis = 0;
    if ( i == FIELD_FILE ) default_vis = 0;
    if ( i == FIELD_PATH ) default_vis = 0;
    if ( i == FIELD_SIZE ) default_vis = 0;

    ColumnVisibility[i] = settings.readNumEntry( 
	QString( "/madman/column.%1.visible" ).arg( i ), 
	default_vis ) != 0;
  }

  READ_BOOL( "http_daemon_enabled", HttpDaemonEnabled, 1 );
  HttpDaemonPort = settings.readNumEntry( "/madman/http_server_port", 51533 );
  READ_BOOL( "http_downloads_enabled", HttpDownloadsEnabled, 1 );
  READ_BOOL( "http_scripting_enabled", HttpScriptingEnabled, 1 );
  READ_BOOL( "http_write_scripting_enabled", HttpWriteScriptingEnabled, 1 );
  READ_BOOL( "http_local_play_enabled", HttpLocalPlayEnabled, 1 );
  READ_BOOL( "http_browsing_enabled", HttpBrowsingEnabled, 1 );

  AutoDJPreferences.PlayRatioUnrated= 
    settings.readNumEntry( "/madman/auto_dj/play_ratio_unrated", 10 );
  AutoDJPreferences.PlayRatio5Rated= 
    settings.readNumEntry( "/madman/auto_dj/play_ratio_5_rated", 10 );
  AutoDJPreferences.PlayRatio4Rated= 
    settings.readNumEntry( "/madman/auto_dj/play_ratio_4_rated", 25 );
  AutoDJPreferences.PlayRatio3Rated= 
    settings.readNumEntry( "/madman/auto_dj/play_ratio_3_rated", 10 );
  AutoDJPreferences.PlayRatio2Rated= 
    settings.readNumEntry( "/madman/auto_dj/play_ratio_2_rated", 5 );
  AutoDJPreferences.PlayRatio1Rated= 
    settings.readNumEntry( "/madman/auto_dj/play_ratio_1_rated", 2 );
  AutoDJPreferences.PlayRatio0Rated= 
    settings.readNumEntry( "/madman/auto_dj/play_ratio_0_rated", 2 );
  AutoDJPreferences.PlayRatioUnplayed= 
    settings.readNumEntry( "/madman/auto_dj/play_ratio_unplayed", 15 );
  AutoDJPreferences.PlayRatioOftenPlayed= 
    settings.readNumEntry( "/madman/auto_dj/play_ratio_often_played", 20 );
  AutoDJPreferences.PlayRatioRarelyPlayed= 
    settings.readNumEntry( "/madman/auto_dj/play_ratio_rarely_played", 12 );
  AutoDJPreferences.PlayRatioRandom= 
    settings.readNumEntry( "/madman/auto_dj/play_ratio_random", 5 );
  AutoDJPreferences.PlayRatioUnratedByGoodArtist = 
    settings.readNumEntry( "/madman/auto_dj/play_ratio_unrated_by_good_artist", 25 );
}




void tPreferences::setPlayer( tPlayer *newplayer )
{
  if ( newplayer )
  {
    if ( Player ) 
      delete Player;
    Player = newplayer;
  }
}





